from __future__ import annotations

from ..providers.types          import BaseProvider, ProviderType
from ..providers.retry_provider import RetryProvider, IterListProvider
from ..providers.base_provider  import AsyncProvider, AsyncGeneratorProvider
from ..providers.create_images  import CreateImagesProvider
from .. import debug
try:
    from .deprecated import *
except ImportError as e:
    debug.error("Deprecated providers not loaded:", e)
from .needs_auth       import *
from .template         import OpenaiTemplate, BackendApi, Puter
from .hf               import HuggingFace, HuggingChat, HuggingFaceAPI, HuggingFaceInference, HuggingFaceMedia
from .har              import HarProvider
try:
    from .not_working import *
except ImportError as e:
    debug.error("Not working providers not loaded:", e)
try:
    from .local import *
except ImportError as e:
    debug.error("Local providers not loaded:", e)
try:
    from .hf_space import *
except ImportError as e:
    debug.error("HuggingFace Space providers not loaded:", e)
try:
    from .mini_max import HailuoAI, MiniMax
except ImportError as e:
    debug.error("MiniMax providers not loaded:", e)
try:
    from .audio import *
except ImportError as e:
    debug.error("Audio providers not loaded:", e)

from .ARTA                 import ARTA
from .Blackbox             import Blackbox
from .Chatai               import Chatai
from .ChatGLM              import ChatGLM
from .ChatGpt              import ChatGpt
from .Cloudflare           import Cloudflare
from .Copilot              import Copilot
from .DDG                  import DDG
from .DeepInfraChat        import DeepInfraChat
from .DocsBot              import DocsBot
from .DuckDuckGo           import DuckDuckGo
from .Dynaspark            import Dynaspark
from .Free2GPT             import Free2GPT
from .FreeGpt              import FreeGpt
from .GizAI                import GizAI
from .ImageLabs            import ImageLabs
from .LambdaChat           import LambdaChat
from .Liaobots             import Liaobots
from .LMArenaProvider      import LMArenaProvider
from .OIVSCodeSer2         import OIVSCodeSer2
from .OIVSCodeSer5         import OIVSCodeSer5
from .OIVSCodeSer0501      import OIVSCodeSer0501
from .PerplexityLabs       import PerplexityLabs
from .Pi                   import Pi
from .Pizzagpt             import Pizzagpt
from .PuterJS              import PuterJS
from .PollinationsAI       import PollinationsAI
from .PollinationsImage    import PollinationsImage
from .TeachAnything        import TeachAnything
from .You                  import You
from .Websim               import Websim
from .WeWordle             import WeWordle
from .Yqcloud              import Yqcloud

import sys

__modules__: list = [
    getattr(sys.modules[__name__], provider) for provider in dir()
    if not provider.startswith("__")
]
__providers__: list[ProviderType] = [
    provider for provider in __modules__
    if isinstance(provider, type)
    and issubclass(provider, BaseProvider)
]
__all__: list[str] = [
    provider.__name__ for provider in __providers__
]
__map__: dict[str, ProviderType] = {
    provider.__name__: provider for provider in __providers__
}

class ProviderUtils:
    convert: dict[str, ProviderType] = __map__



from __future__ import annotations

from aiohttp import ClientSession
import json
import time
import re
import random
import json # Ensure json is imported if not already
import uuid # Import the uuid module
import asyncio
from typing import Optional, Dict, Any, List, Union # Assuming these are used elsewhere or can be pruned if not

from ..typing import AsyncResult, Messages, MediaListType
from .base_provider import AsyncGeneratorProvider, ProviderModelMixin
from ..providers.response import FinishReason, JsonConversation
from ..tools.media import merge_media
from ..image import to_data_uri, is_data_uri_an_image
from ..errors import RateLimitError, ModelNotFoundError
from .. import debug

class AuthData:
    """
    Stores authentication data for a specific model.
    """
    def __init__(self):
        self.auth_token: Optional[str] = None
        self.app_token: Optional[str] = None
        self.created_at: float = time.time()
        self.tokens_valid: bool = False
        self.rate_limited_until: float = 0
        
    def is_valid(self, expiration_time: int) -> bool:
        """Check if the tokens are still valid based on expiration time."""
        return (self.auth_token and self.app_token and 
                self.tokens_valid and 
                time.time() - self.created_at < expiration_time)
    
    def invalidate(self):
        """Mark tokens as invalid."""
        self.tokens_valid = False
    
    def set_rate_limit(self, seconds: int = 60):
        """Set rate limit expiry time."""
        self.rate_limited_until = time.time() + seconds
    
    def is_rate_limited(self) -> bool:
        """Check if currently rate limited."""
        return time.time() < self.rate_limited_until


class Conversation(JsonConversation):
    """
    Stores conversation state and authentication tokens for PuterJS provider.
    Maintains separate authentication data for different models.
    """
    message_history: Messages = []
    
    def __init__(self, model: str):
        self.model = model
        self.message_history = []
        # Authentication tokens by specific model
        self._auth_data: Dict[str, AuthData] = {}
    
    def get_auth_for_model(self, model: str, provider: PuterJS) -> AuthData:
        """Get the authentication data for a specific model."""
        # Create auth data for this model if it doesn't exist
        if model not in self._auth_data:
            self._auth_data[model] = AuthData()
        
        return self._auth_data[model]
    
    # Override get_dict to exclude auth_data
    def get_dict(self) -> Dict[str, Any]:
        """Return a dictionary representation for JSON serialization."""
        return {
            "model": self.model,
            "message_history": self.message_history
        }
    
    # Override __getstate__ for pickle serialization
    def __getstate__(self) -> Dict[str, Any]:
        """Support for pickle serialization."""
        # Include auth_data for pickle serialization
        state = self.__dict__.copy()
        return state
    
    # Override __setstate__ for pickle deserialization
    def __setstate__(self, state: Dict[str, Any]) -> None:
        """Support for pickle deserialization."""
        self.__dict__.update(state)


class PuterJS(AsyncGeneratorProvider, ProviderModelMixin):
    label = "Puter.js"
    url = "https://docs.puter.com/playground"
    api_endpoint = "https://api.puter.com/drivers/call"
    
    working = True
    needs_auth = False
    supports_stream = True
    supports_system_message = True
    supports_message_history = True
    
    default_model = 'gpt-4o'
    default_vision_model = default_model
    # https://api.puter.com/puterai/chat/models/
    openai_models = [default_vision_model,"gpt-4o-mini", "o1", "o1-mini", "o1-pro", "o3", "o3-mini", "o4-mini", "gpt-4.1", "gpt-4.1-mini", "gpt-4.1-nano", "gpt-4.5-preview"]
    claude_models = ["claude-3-7-sonnet-20250219", "claude-3-7-sonnet-latest", "claude-3-5-sonnet-20241022", "claude-3-5-sonnet-latest", "claude-3-5-sonnet-20240620", "claude-3-haiku-20240307"]
    mistral_models = ["ministral-3b-2410","ministral-3b-latest","ministral-8b-2410","ministral-8b-latest","open-mistral-7b","mistral-tiny","mistral-tiny-2312","open-mixtral-8x7b","mistral-small","mistral-small-2312","open-mixtral-8x22b","open-mixtral-8x22b-2404","mistral-large-2411","mistral-large-latest","pixtral-large-2411","pixtral-large-latest","mistral-large-pixtral-2411","codestral-2501","codestral-latest","codestral-2412","codestral-2411-rc5","pixtral-12b-2409","pixtral-12b","pixtral-12b-latest","mistral-small-2503","mistral-small-latest"]
    xai_models = ["grok-beta", "grok-vision-beta"]
    deepseek_models = ["deepseek-chat","deepseek-reasoner"]
    gemini_models = ["gemini-1.5-flash","gemini-2.0-flash"]
    openrouter_models = ["openrouter:meta-llama/llama-3.3-8b-instruct:free","openrouter:nousresearch/deephermes-3-mistral-24b-preview:free","openrouter:mistralai/mistral-medium-3","openrouter:google/gemini-2.5-pro-preview","openrouter:arcee-ai/caller-large","openrouter:arcee-ai/spotlight","openrouter:arcee-ai/maestro-reasoning","openrouter:arcee-ai/virtuoso-large","openrouter:arcee-ai/coder-large","openrouter:arcee-ai/virtuoso-medium-v2","openrouter:arcee-ai/arcee-blitz","openrouter:microsoft/phi-4-reasoning-plus:free","openrouter:microsoft/phi-4-reasoning-plus","openrouter:microsoft/phi-4-reasoning:free","openrouter:qwen/qwen3-0.6b-04-28:free","openrouter:inception/mercury-coder-small-beta","openrouter:qwen/qwen3-1.7b:free","openrouter:qwen/qwen3-4b:free","openrouter:opengvlab/internvl3-14b:free","openrouter:opengvlab/internvl3-2b:free","openrouter:deepseek/deepseek-prover-v2:free","openrouter:deepseek/deepseek-prover-v2","openrouter:meta-llama/llama-guard-4-12b","openrouter:qwen/qwen3-30b-a3b:free","openrouter:qwen/qwen3-30b-a3b","openrouter:qwen/qwen3-8b:free","openrouter:qwen/qwen3-8b","openrouter:qwen/qwen3-14b:free","openrouter:qwen/qwen3-14b","openrouter:qwen/qwen3-32b:free","openrouter:qwen/qwen3-32b","openrouter:qwen/qwen3-235b-a22b:free","openrouter:qwen/qwen3-235b-a22b","openrouter:tngtech/deepseek-r1t-chimera:free","openrouter:thudm/glm-z1-rumination-32b","openrouter:thudm/glm-z1-9b:free","openrouter:thudm/glm-4-9b:free","openrouter:microsoft/mai-ds-r1:free","openrouter:thudm/glm-z1-32b:free","openrouter:thudm/glm-z1-32b","openrouter:thudm/glm-4-32b:free","openrouter:thudm/glm-4-32b","openrouter:google/gemini-2.5-flash-preview","openrouter:google/gemini-2.5-flash-preview:thinking","openrouter:openai/o4-mini-high","openrouter:openai/o3","openrouter:openai/o4-mini","openrouter:shisa-ai/shisa-v2-llama3.3-70b:free","openrouter:qwen/qwen2.5-coder-7b-instruct","openrouter:openai/gpt-4.1","openrouter:openai/gpt-4.1-mini","openrouter:openai/gpt-4.1-nano","openrouter:eleutherai/llemma_7b","openrouter:alfredpros/codellama-7b-instruct-solidity","openrouter:arliai/qwq-32b-arliai-rpr-v1:free","openrouter:agentica-org/deepcoder-14b-preview:free","openrouter:moonshotai/kimi-vl-a3b-thinking:free","openrouter:x-ai/grok-3-mini-beta","openrouter:x-ai/grok-3-beta","openrouter:nvidia/llama-3.3-nemotron-super-49b-v1:free","openrouter:nvidia/llama-3.3-nemotron-super-49b-v1","openrouter:nvidia/llama-3.1-nemotron-ultra-253b-v1:free","openrouter:meta-llama/llama-4-maverick:free","openrouter:meta-llama/llama-4-maverick","openrouter:meta-llama/llama-4-scout:free","openrouter:meta-llama/llama-4-scout","openrouter:all-hands/openhands-lm-32b-v0.1","openrouter:mistral/ministral-8b","openrouter:deepseek/deepseek-v3-base:free","openrouter:scb10x/llama3.1-typhoon2-8b-instruct","openrouter:scb10x/llama3.1-typhoon2-70b-instruct","openrouter:bytedance-research/ui-tars-72b:free","openrouter:qwen/qwen2.5-vl-3b-instruct:free","openrouter:google/gemini-2.5-pro-exp-03-25","openrouter:qwen/qwen2.5-vl-32b-instruct:free","openrouter:qwen/qwen2.5-vl-32b-instruct","openrouter:deepseek/deepseek-chat-v3-0324:free","openrouter:deepseek/deepseek-chat-v3-0324","openrouter:featherless/qwerky-72b:free","openrouter:openai/o1-pro","openrouter:mistralai/mistral-small-3.1-24b-instruct:free","openrouter:mistralai/mistral-small-3.1-24b-instruct","openrouter:open-r1/olympiccoder-32b:free","openrouter:google/gemma-3-1b-it:free","openrouter:google/gemma-3-4b-it:free","openrouter:google/gemma-3-4b-it","openrouter:ai21/jamba-1.6-large","openrouter:ai21/jamba-1.6-mini","openrouter:google/gemma-3-12b-it:free","openrouter:google/gemma-3-12b-it","openrouter:cohere/command-a","openrouter:openai/gpt-4o-mini-search-preview","openrouter:openai/gpt-4o-search-preview","openrouter:rekaai/reka-flash-3:free","openrouter:google/gemma-3-27b-it:free","openrouter:google/gemma-3-27b-it","openrouter:thedrummer/anubis-pro-105b-v1","openrouter:thedrummer/skyfall-36b-v2","openrouter:microsoft/phi-4-multimodal-instruct","openrouter:perplexity/sonar-reasoning-pro","openrouter:perplexity/sonar-pro","openrouter:perplexity/sonar-deep-research","openrouter:deepseek/deepseek-r1-zero:free","openrouter:qwen/qwq-32b:free","openrouter:qwen/qwq-32b","openrouter:moonshotai/moonlight-16b-a3b-instruct:free","openrouter:nousresearch/deephermes-3-llama-3-8b-preview:free","openrouter:openai/gpt-4.5-preview","openrouter:google/gemini-2.0-flash-lite-001","openrouter:anthropic/claude-3.7-sonnet","openrouter:anthropic/claude-3.7-sonnet:thinking","openrouter:anthropic/claude-3.7-sonnet:beta","openrouter:perplexity/r1-1776","openrouter:mistralai/mistral-saba","openrouter:cognitivecomputations/dolphin3.0-r1-mistral-24b:free","openrouter:cognitivecomputations/dolphin3.0-mistral-24b:free","openrouter:meta-llama/llama-guard-3-8b","openrouter:openai/o3-mini-high","openrouter:deepseek/deepseek-r1-distill-llama-8b","openrouter:google/gemini-2.0-flash-001","openrouter:qwen/qwen-vl-plus","openrouter:aion-labs/aion-1.0","openrouter:aion-labs/aion-1.0-mini","openrouter:aion-labs/aion-rp-llama-3.1-8b","openrouter:qwen/qwen-vl-max","openrouter:qwen/qwen-turbo","openrouter:qwen/qwen2.5-vl-72b-instruct:free","openrouter:qwen/qwen2.5-vl-72b-instruct","openrouter:qwen/qwen-plus","openrouter:qwen/qwen-max","openrouter:openai/o3-mini","openrouter:deepseek/deepseek-r1-distill-qwen-1.5b","openrouter:mistralai/mistral-small-24b-instruct-2501:free","openrouter:mistralai/mistral-small-24b-instruct-2501","openrouter:deepseek/deepseek-r1-distill-qwen-32b:free","openrouter:deepseek/deepseek-r1-distill-qwen-32b","openrouter:deepseek/deepseek-r1-distill-qwen-14b:free","openrouter:deepseek/deepseek-r1-distill-qwen-14b","openrouter:perplexity/sonar-reasoning","openrouter:perplexity/sonar","openrouter:liquid/lfm-7b","openrouter:liquid/lfm-3b","openrouter:deepseek/deepseek-r1-distill-llama-70b:free","openrouter:deepseek/deepseek-r1-distill-llama-70b","openrouter:deepseek/deepseek-r1:free","openrouter:deepseek/deepseek-r1","openrouter:minimax/minimax-01","openrouter:mistralai/codestral-2501","openrouter:microsoft/phi-4","openrouter:deepseek/deepseek-chat:free","openrouter:deepseek/deepseek-chat","openrouter:sao10k/l3.3-euryale-70b","openrouter:openai/o1","openrouter:eva-unit-01/eva-llama-3.33-70b","openrouter:x-ai/grok-2-vision-1212","openrouter:x-ai/grok-2-1212","openrouter:cohere/command-r7b-12-2024","openrouter:google/gemini-2.0-flash-exp:free","openrouter:meta-llama/llama-3.3-70b-instruct:free","openrouter:meta-llama/llama-3.3-70b-instruct","openrouter:amazon/nova-lite-v1","openrouter:amazon/nova-micro-v1","openrouter:amazon/nova-pro-v1","openrouter:qwen/qwq-32b-preview","openrouter:eva-unit-01/eva-qwen-2.5-72b","openrouter:openai/gpt-4o-2024-11-20","openrouter:mistralai/mistral-large-2411","openrouter:mistralai/mistral-large-2407","openrouter:mistralai/pixtral-large-2411","openrouter:x-ai/grok-vision-beta","openrouter:infermatic/mn-inferor-12b","openrouter:qwen/qwen-2.5-coder-32b-instruct:free","openrouter:qwen/qwen-2.5-coder-32b-instruct","openrouter:raifle/sorcererlm-8x22b","openrouter:eva-unit-01/eva-qwen-2.5-32b","openrouter:thedrummer/unslopnemo-12b","openrouter:anthropic/claude-3.5-haiku:beta","openrouter:anthropic/claude-3.5-haiku","openrouter:anthropic/claude-3.5-haiku-20241022:beta","openrouter:anthropic/claude-3.5-haiku-20241022","openrouter:neversleep/llama-3.1-lumimaid-70b","openrouter:anthracite-org/magnum-v4-72b","openrouter:anthropic/claude-3.5-sonnet:beta","openrouter:anthropic/claude-3.5-sonnet","openrouter:x-ai/grok-beta","openrouter:mistralai/ministral-8b","openrouter:mistralai/ministral-3b","openrouter:qwen/qwen-2.5-7b-instruct:free","openrouter:qwen/qwen-2.5-7b-instruct","openrouter:nvidia/llama-3.1-nemotron-70b-instruct","openrouter:inflection/inflection-3-productivity","openrouter:inflection/inflection-3-pi","openrouter:google/gemini-flash-1.5-8b","openrouter:thedrummer/rocinante-12b","openrouter:anthracite-org/magnum-v2-72b","openrouter:liquid/lfm-40b","openrouter:meta-llama/llama-3.2-3b-instruct:free","openrouter:meta-llama/llama-3.2-3b-instruct","openrouter:meta-llama/llama-3.2-1b-instruct:free","openrouter:meta-llama/llama-3.2-1b-instruct","openrouter:meta-llama/llama-3.2-90b-vision-instruct","openrouter:meta-llama/llama-3.2-11b-vision-instruct:free","openrouter:meta-llama/llama-3.2-11b-vision-instruct","openrouter:qwen/qwen-2.5-72b-instruct:free","openrouter:qwen/qwen-2.5-72b-instruct","openrouter:neversleep/llama-3.1-lumimaid-8b","openrouter:openai/o1-preview","openrouter:openai/o1-preview-2024-09-12","openrouter:openai/o1-mini","openrouter:openai/o1-mini-2024-09-12","openrouter:mistralai/pixtral-12b","openrouter:cohere/command-r-plus-08-2024","openrouter:cohere/command-r-08-2024","openrouter:qwen/qwen-2.5-vl-7b-instruct:free","openrouter:qwen/qwen-2.5-vl-7b-instruct","openrouter:sao10k/l3.1-euryale-70b","openrouter:microsoft/phi-3.5-mini-128k-instruct","openrouter:nousresearch/hermes-3-llama-3.1-70b","openrouter:nousresearch/hermes-3-llama-3.1-405b","openrouter:openai/chatgpt-4o-latest","openrouter:sao10k/l3-lunaris-8b","openrouter:aetherwiing/mn-starcannon-12b","openrouter:openai/gpt-4o-2024-08-06","openrouter:meta-llama/llama-3.1-405b:free","openrouter:meta-llama/llama-3.1-405b","openrouter:nothingiisreal/mn-celeste-12b","openrouter:perplexity/llama-3.1-sonar-small-128k-online","openrouter:perplexity/llama-3.1-sonar-large-128k-online","openrouter:meta-llama/llama-3.1-8b-instruct:free","openrouter:meta-llama/llama-3.1-8b-instruct","openrouter:meta-llama/llama-3.1-405b-instruct","openrouter:meta-llama/llama-3.1-70b-instruct","openrouter:mistralai/codestral-mamba","openrouter:mistralai/mistral-nemo:free","openrouter:mistralai/mistral-nemo","openrouter:openai/gpt-4o-mini","openrouter:openai/gpt-4o-mini-2024-07-18","openrouter:google/gemma-2-27b-it","openrouter:alpindale/magnum-72b","openrouter:google/gemma-2-9b-it:free","openrouter:google/gemma-2-9b-it","openrouter:01-ai/yi-large","openrouter:ai21/jamba-instruct","openrouter:anthropic/claude-3.5-sonnet-20240620:beta","openrouter:anthropic/claude-3.5-sonnet-20240620","openrouter:sao10k/l3-euryale-70b","openrouter:cognitivecomputations/dolphin-mixtral-8x22b","openrouter:qwen/qwen-2-72b-instruct","openrouter:mistralai/mistral-7b-instruct:free","openrouter:mistralai/mistral-7b-instruct","openrouter:nousresearch/hermes-2-pro-llama-3-8b","openrouter:mistralai/mistral-7b-instruct-v0.3","openrouter:microsoft/phi-3-mini-128k-instruct","openrouter:microsoft/phi-3-medium-128k-instruct","openrouter:neversleep/llama-3-lumimaid-70b","openrouter:deepseek/deepseek-coder","openrouter:google/gemini-flash-1.5","openrouter:openai/gpt-4o","openrouter:openai/gpt-4o:extended","openrouter:meta-llama/llama-guard-2-8b","openrouter:openai/gpt-4o-2024-05-13","openrouter:allenai/olmo-7b-instruct","openrouter:neversleep/llama-3-lumimaid-8b:extended","openrouter:neversleep/llama-3-lumimaid-8b","openrouter:sao10k/fimbulvetr-11b-v2","openrouter:meta-llama/llama-3-8b-instruct","openrouter:meta-llama/llama-3-70b-instruct","openrouter:mistralai/mixtral-8x22b-instruct","openrouter:microsoft/wizardlm-2-8x22b","openrouter:google/gemini-pro-1.5","openrouter:openai/gpt-4-turbo","openrouter:cohere/command-r-plus","openrouter:cohere/command-r-plus-04-2024","openrouter:sophosympatheia/midnight-rose-70b","openrouter:cohere/command","openrouter:cohere/command-r","openrouter:anthropic/claude-3-haiku:beta","openrouter:anthropic/claude-3-haiku","openrouter:anthropic/claude-3-opus:beta","openrouter:anthropic/claude-3-opus","openrouter:anthropic/claude-3-sonnet:beta","openrouter:anthropic/claude-3-sonnet","openrouter:cohere/command-r-03-2024","openrouter:mistralai/mistral-large","openrouter:openai/gpt-3.5-turbo-0613","openrouter:openai/gpt-4-turbo-preview","openrouter:nousresearch/nous-hermes-2-mixtral-8x7b-dpo","openrouter:mistralai/mistral-medium","openrouter:mistralai/mistral-small","openrouter:mistralai/mistral-tiny","openrouter:mistralai/mistral-7b-instruct-v0.2","openrouter:mistralai/mixtral-8x7b-instruct","openrouter:neversleep/noromaid-20b","openrouter:anthropic/claude-2.1:beta","openrouter:anthropic/claude-2.1","openrouter:anthropic/claude-2:beta","openrouter:anthropic/claude-2","openrouter:undi95/toppy-m-7b","openrouter:alpindale/goliath-120b","openrouter:openrouter/auto","openrouter:openai/gpt-3.5-turbo-1106","openrouter:openai/gpt-4-1106-preview","openrouter:jondurbin/airoboros-l2-70b","openrouter:openai/gpt-3.5-turbo-instruct","openrouter:mistralai/mistral-7b-instruct-v0.1","openrouter:pygmalionai/mythalion-13b","openrouter:openai/gpt-3.5-turbo-16k","openrouter:openai/gpt-4-32k","openrouter:openai/gpt-4-32k-0314","openrouter:mancer/weaver","openrouter:anthropic/claude-2.0:beta","openrouter:anthropic/claude-2.0","openrouter:undi95/remm-slerp-l2-13b","openrouter:gryphe/mythomax-l2-13b","openrouter:meta-llama/llama-2-70b-chat","openrouter:openai/gpt-3.5-turbo","openrouter:openai/gpt-3.5-turbo-0125","openrouter:openai/gpt-4","openrouter:openai/gpt-4-0314"]
    
    vision_models = [*openai_models, *claude_models, *mistral_models, *xai_models, *deepseek_models, *gemini_models, *openrouter_models]
    
    models = vision_models
    
    model_aliases = {              
        ### mistral_models ###
        "mixtral-8x22b": ["open-mixtral-8x22b", "open-mixtral-8x22b-2404"],
        "pixtral-large": ["pixtral-large-2411","pixtral-large-latest", "mistral-large-pixtral-2411"],
        
        ### openrouter_models ###
        # llama
        "llama-2-70b": "openrouter:meta-llama/llama-2-70b-chat",
        "llama-3-8b": "openrouter:meta-llama/llama-3-8b-instruct",
        "llama-3-70b": "openrouter:meta-llama/llama-3-70b-instruct",
        "llama-3.1-8b": ["openrouter:meta-llama/llama-3.1-8b-instruct:free", "openrouter:meta-llama/llama-3.1-8b-instruct"],
        "llama-3.1-70b": "openrouter:meta-llama/llama-3.1-70b-instruct",
        "llama-3.1-405b": ["openrouter:meta-llama/llama-3.1-405b:free", "openrouter:meta-llama/llama-3.1-405b", "openrouter:meta-llama/llama-3.1-405b-instruct"],
        "llama-3.2-1b": ["openrouter:meta-llama/llama-3.2-1b-instruct:free", "openrouter:meta-llama/llama-3.2-1b-instruct"],
        "llama-3.2-3b": ["openrouter:meta-llama/llama-3.2-3b-instruct:free","openrouter:meta-llama/llama-3.2-3b-instruct"],
        "llama-3.2-11b": ["openrouter:meta-llama/llama-3.2-11b-vision-instruct:free", "openrouter:meta-llama/llama-3.2-11b-vision-instruct"],
        "llama-3.2-90b": "openrouter:meta-llama/llama-3.2-90b-vision-instruct",
        "llama-3.3-8b": "openrouter:meta-llama/llama-3.3-8b-instruct:free",
        "llama-3.3-70b": ["openrouter:meta-llama/llama-3.3-70b-instruct:free", "openrouter:meta-llama/llama-3.3-70b-instruct"],
        "llama-4-maverick": ["openrouter:meta-llama/llama-4-maverick:free", "openrouter:meta-llama/llama-4-maverick"],
        "llama-4-scout": ["openrouter:meta-llama/llama-4-scout:free", "openrouter:meta-llama/llama-4-scout"],
        #"": "openrouter:meta-llama/llama-guard-3-8b",
        #"": "openrouter:meta-llama/llama-guard-2-8b",
        #"": "openrouter:meta-llama/llama-guard-4-12b",
        
        # google (gemini)
        "gemini-1.5-flash": ["gemini-1.5-flash", "openrouter:google/gemini-flash-1.5", "gemini-flash-1.5-8b"],
        "gemini-1.5-8b-flash": "openrouter:google/gemini-flash-1.5-8b",
        "gemini-1.5-pro": "openrouter:google/gemini-pro-1.5",
        "gemini-2.0-flash": ["gemini-2.0-flash", "openrouter:google/gemini-2.0-flash-lite-001", "openrouter:google/gemini-2.0-flash-001", "openrouter:google/gemini-2.0-flash-exp:free"],
        "gemini-2.5-pro": ["openrouter:google/gemini-2.5-pro-preview", "openrouter:google/gemini-2.5-pro-exp-03-25"],
        "gemini-2.5-flash": "openrouter:google/gemini-2.5-flash-preview",
        "gemini-2.5-flash-thinking": "openrouter:google/gemini-2.5-flash-preview:thinking",
        
        # google (gemma)
        "gemma-2-9b": ["openrouter:google/gemma-2-9b-it:free","openrouter:google/gemma-2-9b-it"],
        "gemma-2-27b": "openrouter:google/gemma-2-27b-it",
        "gemma-3-1b": "openrouter:google/gemma-3-1b-it:free",
        "gemma-3-4b": ["openrouter:google/gemma-3-4b-it:free", "openrouter:google/gemma-3-4b-it"],
        "gemma-3-12b": ["openrouter:google/gemma-3-12b-it:free", "openrouter:google/gemma-3-12b-it"],
        "gemma-3-27b": ["openrouter:google/gemma-3-27b-it:free", "openrouter:google/gemma-3-27b-it"],
        
        # openai (gpt-3.5)
        "gpt-3.5-turbo": ["openrouter:openai/gpt-3.5-turbo-0613", "openrouter:openai/gpt-3.5-turbo-1106", "openrouter:openai/gpt-3.5-turbo-0125", "openrouter:openai/gpt-3.5-turbo", "openrouter:openai/gpt-3.5-turbo-instruct", "openrouter:openai/gpt-3.5-turbo-16k"],
        
        # openai (gpt-4)
        "gpt-4": ["openrouter:openai/gpt-4-1106-preview", "openrouter:openai/gpt-4-32k", "openrouter:openai/gpt-4-32k-0314", "openrouter:openai/gpt-4", "openrouter:openai/gpt-4-0314",],
        "gpt-4-turbo": ["openrouter:openai/gpt-4-turbo", "openrouter:openai/gpt-4-turbo-preview"],
        
        # openai (gpt-4o)
        "gpt-4o": ["gpt-4o", "openrouter:openai/gpt-4o-2024-08-06", "openrouter:openai/gpt-4o-2024-11-20", "openrouter:openai/chatgpt-4o-latest", "openrouter:openai/gpt-4o", "openrouter:openai/gpt-4o:extended", "openrouter:openai/gpt-4o-2024-05-13",],
        "gpt-4o-search": "openrouter:openai/gpt-4o-search-preview",
        "gpt-4o-mini": ["gpt-4o-mini", "openrouter:openai/gpt-4o-mini",  "openrouter:openai/gpt-4o-mini-2024-07-18"],
        "gpt-4o-mini-search": "openrouter:openai/gpt-4o-mini-search-preview",
        
        # openai (o1)
        "o1": ["o1", "openrouter:openai/o1", "openrouter:openai/o1-preview", "openrouter:openai/o1-preview-2024-09-12"],
        "o1-mini": ["o1-mini", "openrouter:openai/o1-mini", "openrouter:openai/o1-mini-2024-09-12"],
        "o1-pro": ["o1-pro", "openrouter:openai/o1-pro"],
        
        # openai (o3)
        "o3": ["o3", "openrouter:openai/o3"],
        "o3-mini": ["o3-mini", "openrouter:openai/o3-mini", "openrouter:openai/o3-mini-high"],
        "o3-mini-high": "openrouter:openai/o3-mini-high",
        
        # openai (o4)
        "o4-mini": ["o4-mini", "openrouter:openai/o4-mini"],
        "o4-mini-high": "openrouter:openai/o4-mini-high",
        
        # openai (gpt-4.1)
        "gpt-4.1": ["gpt-4.1", "openrouter:openai/gpt-4.1"],
        "gpt-4.1-mini": ["gpt-4.1-mini", "openrouter:openai/gpt-4.1-mini"],
        "gpt-4.1-nano": ["gpt-4.1-nano", "openrouter:openai/gpt-4.1-nano"],
        
        # openai (gpt-4.5)
        "gpt-4.5": ["gpt-4.5-preview", "openrouter:openai/gpt-4.5-preview"],
        
        # mistralai
        "mistral-large": ["openrouter:mistralai/mistral-large", "openrouter:mistralai/mistral-large-2411", "openrouter:mistralai/mistral-large-2407", "openrouter:mistralai/pixtral-large-2411"], 
        "mistral-medium": ["openrouter:mistralai/mistral-medium", "openrouter:mistralai/mistral-medium-3"], 
        "mistral-small": ["mistral-small", "mistral-small-2312", "mistral-small-2503","mistral-small-latest", "openrouter:mistralai/mistral-small", "openrouter:mistralai/mistral-small-3.1-24b-instruct:free", "openrouter:mistralai/mistral-small-3.1-24b-instruct", "openrouter:mistralai/mistral-small-24b-instruct-2501:free", "openrouter:mistralai/mistral-small-24b-instruct-2501"], 
        "mistral-tiny": ["mistral-tiny", "mistral-tiny-2312", "openrouter:mistralai/mistral-tiny"], 
        "mistral-7b": ["open-mistral-7b", "openrouter:mistralai/mistral-7b-instruct", "openrouter:mistralai/mistral-7b-instruct:free", "openrouter:mistralai/mistral-7b-instruct-v0.1", "openrouter:mistralai/mistral-7b-instruct-v0.2", "openrouter:mistralai/mistral-7b-instruct-v0.3",], 
        "mixtral-8x7b": ["open-mixtral-8x7b", "openrouter:mistralai/mixtral-8x7b-instruct"], 
        "mixtral-8x22b": ["open-mixtral-8x22b", "open-mixtral-8x22b-2404", "openrouter:mistralai/mixtral-8x7b-instruct", "openrouter:mistralai/mixtral-8x22b-instruct"], 
        "ministral-8b": ["ministral-8b-2410", "ministral-8b-latest", "openrouter:mistral/ministral-8b", "openrouter:mistralai/ministral-8b"],
        "mistral-nemo": ["openrouter:mistralai/mistral-nemo:free", "openrouter:mistralai/mistral-nemo"],
        "ministral-3b": ["ministral-3b-2410", "ministral-3b-latest", "openrouter:mistralai/ministral-3b"],
        "mistral-saba": "openrouter:mistralai/mistral-saba",
        "codestral": ["codestral-2501","codestral-latest","codestral-2412","codestral-2411-rc5", "openrouter:mistralai/codestral-2501", "openrouter:mistralai/codestral-mamba"],
        "pixtral-12b": ["pixtral-12b-2409","pixtral-12b","pixtral-12b-latest", "openrouter:mistralai/pixtral-12b"],
        
        # nousresearch
        "hermes-2-dpo": "openrouter:nousresearch/nous-hermes-2-mixtral-8x7b-dpo",
        "hermes-2-pro": "openrouter:nousresearch/hermes-2-pro-llama-3-8b",
        "hermes-3-70b": "openrouter:nousresearch/hermes-3-llama-3.1-70b",
        "hermes-3-405b": "openrouter:nousresearch/hermes-3-llama-3.1-405b",
        "deephermes-3-8b": "openrouter:nousresearch/deephermes-3-llama-3-8b-preview:free",
        "deephermes-3-24b": "openrouter:nousresearch/deephermes-3-mistral-24b-preview:free",
        
        # microsoft
        "phi-3-mini": "openrouter:microsoft/phi-3-mini-128k-instruct",
        "phi-3-medium": "openrouter:microsoft/phi-3-medium-128k-instruct",
        "phi-3.5-mini": "openrouter:microsoft/phi-3.5-mini-128k-instruct",
        "phi-4": "openrouter:microsoft/phi-4",
        "phi-4-multimodal": "openrouter:microsoft/phi-4-multimodal-instruct",
        "phi-4-reasoning": "openrouter:microsoft/phi-4-reasoning:free",
        "phi-4-reasoning-plus": ["openrouter:microsoft/phi-4-reasoning-plus:free", "openrouter:microsoft/phi-4-reasoning-plus"],
        
        "wizardlm-2-8x22b": "openrouter:microsoft/wizardlm-2-8x22b",
        
        "mai-ds-r1": "openrouter:microsoft/mai-ds-r1:free",     
        
        # anthropic
        "claude-3.7-sonnet": ["claude-3-7-sonnet-20250219", "claude-3-7-sonnet-latest", "openrouter:anthropic/claude-3.7-sonnet", "openrouter:anthropic/claude-3.7-sonnet:beta",],
        "claude-3.7-sonnet-thinking": "openrouter:anthropic/claude-3.7-sonnet:thinking",
        "claude-3.5-haiku": ["openrouter:anthropic/claude-3.5-haiku:beta", "openrouter:anthropic/claude-3.5-haiku", "openrouter:anthropic/claude-3.5-haiku-20241022:beta", "openrouter:anthropic/claude-3.5-haiku-20241022"],
        "claude-3.5-sonnet": ["claude-3-5-sonnet-20241022", "claude-3-5-sonnet-latest", "claude-3-5-sonnet-20240620", "openrouter:anthropic/claude-3.5-sonnet-20240620:beta", "openrouter:anthropic/claude-3.5-sonnet-20240620", "openrouter:anthropic/claude-3.5-sonnet:beta", "openrouter:anthropic/claude-3.5-sonnet",],
        "claude-3-haiku": ["claude-3-haiku-20240307", "openrouter:anthropic/claude-3-haiku:beta", "openrouter:anthropic/claude-3-haiku"],
        "claude-3-opus": ["openrouter:anthropic/claude-3-opus:beta", "openrouter:anthropic/claude-3-opus"],
        "claude-3-sonnet": ["openrouter:anthropic/claude-3-sonnet:beta", "openrouter:anthropic/claude-3-sonnet"],
        "claude-2.1": ["openrouter:anthropic/claude-2.1:beta", "openrouter:anthropic/claude-2.1"],
        "claude-2": ["openrouter:anthropic/claude-2:beta", "openrouter:anthropic/claude-2",],
        "claude-2.0": ["openrouter:anthropic/claude-2.0:beta", "openrouter:anthropic/claude-2.0"],
        
        # rekaai
        "reka-flash": "openrouter:rekaai/reka-flash-3:free",

        # cohere
        "command-r7b": "openrouter:cohere/command-r7b-12-2024",
        "command-r-plus": ["openrouter:cohere/command-r-plus-08-2024", "openrouter:cohere/command-r-plus", "openrouter:cohere/command-r-plus-04-2024"],
        "command": "openrouter:cohere/command",
        "command-r": ["openrouter:cohere/command-r-08-2024", "openrouter:cohere/command-r", "openrouter:cohere/command-r-03-2024"],
        "command-a": "openrouter:cohere/command-a",
        
        # qwen
        "qwq-32b": ["openrouter:qwen/qwq-32b-preview", "openrouter:qwen/qwq-32b:free", "openrouter:qwen/qwq-32b"],
        "qwen-vl-plus": "openrouter:qwen/qwen-vl-plus",
        "qwen-vl-max": "openrouter:qwen/qwen-vl-max",
        "qwen-turbo": "openrouter:qwen/qwen-turbo",
        "qwen-2.5-vl-72b": ["openrouter:qwen/qwen2.5-vl-72b-instruct:free", "openrouter:qwen/qwen2.5-vl-72b-instruct"],
        "qwen-plus": "openrouter:qwen/qwen-plus",
        "qwen-max": "openrouter:qwen/qwen-max",
        "qwen-2.5-coder-32b": ["openrouter:qwen/qwen-2.5-coder-32b-instruct:free", "openrouter:qwen/qwen-2.5-coder-32b-instruct"],
        "qwen-2.5-7b": ["openrouter:qwen/qwen-2.5-7b-instruct:free", "openrouter:qwen/qwen-2.5-7b-instruct"],
        "qwen-2.5-72b": ["openrouter:qwen/qwen-2.5-72b-instruct:free", "openrouter:qwen/qwen-2.5-72b-instruct"],
        "qwen-2.5-vl-7b": ["openrouter:qwen/qwen-2.5-vl-7b-instruct:free", "openrouter:qwen/qwen-2.5-vl-7b-instruct"],
        "qwen-2-72b": "openrouter:qwen/qwen-2-72b-instruct",
        "qwen-3-0.6b": "openrouter:qwen/qwen3-0.6b-04-28:free",
        "qwen-3-1.7b": "openrouter:qwen/qwen3-1.7b:free",
        "qwen-3-4b": "openrouter:qwen/qwen3-4b:free",
        "qwen-3-30b": ["openrouter:qwen/qwen3-30b-a3b:free", "openrouter:qwen/qwen3-30b-a3b"],
        "qwen-3-8b": ["openrouter:qwen/qwen3-8b:free", "openrouter:qwen/qwen3-8b"],
        "qwen-3-14b": ["openrouter:qwen/qwen3-14b:free", "openrouter:qwen/qwen3-14b"],
        "qwen-3-32b": ["openrouter:qwen/qwen3-32b:free", "openrouter:qwen/qwen3-32b"],
        "qwen-3-235b": ["openrouter:qwen/qwen3-235b-a22b:free", "openrouter:qwen/qwen3-235b-a22b"],
        "qwen-2.5-coder-7b": "openrouter:qwen/qwen2.5-coder-7b-instruct",
        "qwen-2.5-vl-3b": "openrouter:qwen/qwen2.5-vl-3b-instruct:free",
        "qwen-2.5-vl-32b": ["openrouter:qwen/qwen2.5-vl-32b-instruct:free", "openrouter:qwen/qwen2.5-vl-32b-instruct"],
        
        # deepseek
        "deepseek-prover-v2": ["openrouter:deepseek/deepseek-prover-v2:free", "openrouter:deepseek/deepseek-prover-v2"],
        "deepseek-v3": "openrouter:deepseek/deepseek-v3-base:free",
        "deepseek-v3-0324": ["deepseek-chat", "openrouter:deepseek/deepseek-chat-v3-0324:free", "openrouter:deepseek/deepseek-chat-v3-0324"],
        "deepseek-r1-zero": "openrouter:deepseek/deepseek-r1-zero:free",
        "deepseek-r1-distill-llama-8b": "openrouter:deepseek/deepseek-r1-distill-llama-8b",
        "deepseek-r1-distill-qwen-1.5b": "openrouter:deepseek/deepseek-r1-distill-qwen-1.5b",
        "deepseek-r1-distill-qwen-32b": ["openrouter:deepseek/deepseek-r1-distill-qwen-32b:free", "openrouter:deepseek/deepseek-r1-distill-qwen-32b"],
        "deepseek-r1-distill-qwen-14b": ["openrouter:deepseek/deepseek-r1-distill-qwen-14b:free","openrouter:deepseek/deepseek-r1-distill-qwen-14b"],
        "deepseek-r1-distill-llama-70b": ["openrouter:deepseek/deepseek-r1-distill-llama-70b:free", "openrouter:deepseek/deepseek-r1-distill-llama-70b"],
        "deepseek-r1": ["deepseek-reasoner", "openrouter:deepseek/deepseek-r1:free", "openrouter:deepseek/deepseek-r1"],
        "deepseek-chat": ["deepseek-chat", "openrouter:deepseek/deepseek-chat:free", "openrouter:deepseek/deepseek-chat"],
        "deepseek-coder": ["openrouter:deepseek/deepseek-coder"],
        
        # inflection
        "inflection-3-productivity": "openrouter:inflection/inflection-3-productivity",
        "inflection-3-pi": "openrouter:inflection/inflection-3-pi",
        
        # x-ai
        "grok-3-mini": "openrouter:x-ai/grok-3-mini-beta",
        "grok-3-beta": "openrouter:x-ai/grok-3-beta",
        "grok-2": ["openrouter:x-ai/grok-2-vision-1212", "openrouter:x-ai/grok-2-1212"],
        "grok": ["openrouter:x-ai/grok-vision-beta", "openrouter:x-ai/grok-2-vision-1212", "openrouter:x-ai/grok-2-1212", "grok-beta","grok-vision-beta", "openrouter:x-ai/grok-beta", "openrouter:x-ai/grok-3-beta", "openrouter:x-ai/grok-3-mini-beta"],
        "grok-beta": ["grok-beta","grok-vision-beta", "openrouter:x-ai/grok-beta", "openrouter:x-ai/grok-3-beta"],
        
        # perplexity
        "sonar-reasoning-pro": "openrouter:perplexity/sonar-reasoning-pro",
        "sonar-pro": "openrouter:perplexity/sonar-pro",
        "sonar-deep-research": "openrouter:perplexity/sonar-deep-research",
        "r1-1776": "openrouter:perplexity/r1-1776",
        "sonar-reasoning": "openrouter:perplexity/sonar-reasoning",
        "sonar": "openrouter:perplexity/sonar",
        "llama-3.1-sonar-small-online": "openrouter:perplexity/llama-3.1-sonar-small-128k-online",
        "llama-3.1-sonar-large-online": "openrouter:perplexity/llama-3.1-sonar-large-128k-online",
        
        # nvidia
        "nemotron-49b": ["openrouter:nvidia/llama-3.3-nemotron-super-49b-v1:free", "openrouter:nvidia/llama-3.3-nemotron-super-49b-v1"],
        "nemotron-70b": "openrouter:nvidia/llama-3.1-nemotron-70b-instruct",
        "nemotron-253b": "openrouter:nvidia/llama-3.1-nemotron-ultra-253b-v1:free",
        
        # thudm
        "glm-4": ["openrouter:thudm/glm-4-32b:free", "openrouter:thudm/glm-4-32b", "openrouter:thudm/glm-4-9b:free",],
        "glm-4-32b": ["openrouter:thudm/glm-4-32b:free", "openrouter:thudm/glm-4-32b"],
        "glm-z1-32b": ["openrouter:thudm/glm-z1-32b:free", "openrouter:thudm/glm-z1-32b"],
        "glm-4-9b": "openrouter:thudm/glm-4-9b:free",
        "glm-z1-9b": "openrouter:thudm/glm-z1-9b:free",
        "glm-z1-rumination-32b": "openrouter:thudm/glm-z1-rumination-32b",
        
        # minimax
        "minimax": "openrouter:minimax/minimax-01",
        
        # cognitivecomputations
        "dolphin-3.0-r1-24b": "openrouter:cognitivecomputations/dolphin3.0-r1-mistral-24b:free",
        "dolphin-3.0-24b": "openrouter:cognitivecomputations/dolphin3.0-mistral-24b:free",
        "dolphin-8x22b": "openrouter:cognitivecomputations/dolphin-mixtral-8x22b",
        
        # agentica-org
        "deepcoder-14b": "openrouter:agentica-org/deepcoder-14b-preview:free",
        
        # moonshotai
        "kimi-vl-thinking": "openrouter:moonshotai/kimi-vl-a3b-thinking:free",
        "moonlight-16b": "openrouter:moonshotai/moonlight-16b-a3b-instruct:free",
        
        # featherless
        "qwerky-72b": "openrouter:featherless/qwerky-72b:free",
        
        # liquid
        "lfm-7b": "openrouter:liquid/lfm-7b",
        "lfm-3b": "openrouter:liquid/lfm-3b",
        "lfm-40b": "openrouter:liquid/lfm-40b",
        
        # arcee-ai
        #"": "openrouter:arcee-ai/caller-large",
        #"": "openrouter:arcee-ai/spotlight",
        #"": "openrouter:arcee-ai/maestro-reasoning",
        #"": "openrouter:arcee-ai/virtuoso-large",
        #"": "openrouter:arcee-ai/coder-large",
        #"": "openrouter:arcee-ai/virtuoso-medium-v2",
        #"":  "openrouter:arcee-ai/arcee-blitz",
        
        # inception
        #"": "openrouter:inception/mercury-coder-small-beta",

        # opengvlab
        #"": "openrouter:opengvlab/internvl3-14b:free",
        #"": "openrouter:opengvlab/internvl3-2b:free",    
        
        # tngtech
        #"": "openrouter:tngtech/deepseek-r1t-chimera:free",
        
        # shisa-ai
        #"": "openrouter:shisa-ai/shisa-v2-llama3.3-70b:free",
        
        # eleutherai
        #"": "openrouter:eleutherai/llemma_7b",
        
        # shisa-ai
        #"": "openrouter:alfredpros/codellama-7b-instruct-solidity",
        
        # arliai
        #"": "openrouter:arliai/qwq-32b-arliai-rpr-v1:free",
        
        # all-hands
        #"": "openrouter:all-hands/openhands-lm-32b-v0.1",
        
        # scb10x
        #"": "openrouter:scb10x/llama3.1-typhoon2-8b-instruct",
        #"": "openrouter:scb10x/llama3.1-typhoon2-70b-instruct",
        
        # bytedance-research
        #"": "openrouter:bytedance-research/ui-tars-72b:free",
        
        # open-r1
        #"": "openrouter:open-r1/olympiccoder-32b:free",
        
        # ai21
        #"": "openrouter:ai21/jamba-1.6-large",
        #"": "openrouter:ai21/jamba-1.6-mini",
        #"": "openrouter:ai21/jamba-instruct",
        
        # thedrummer
        #"": "openrouter:thedrummer/anubis-pro-105b-v1",
        #"": "openrouter:thedrummer/skyfall-36b-v2",
        #"": "openrouter:thedrummer/unslopnemo-12b",
        #"": "openrouter:thedrummer/rocinante-12b",
        
        # aion-labs
        #"": "openrouter:aion-labs/aion-1.0",
        #"": "openrouter:aion-labs/aion-1.0-mini",
        #"": "openrouter:aion-labs/aion-rp-llama-3.1-8b",
        
        # sao10k
        #"": "openrouter:sao10k/l3.3-euryale-70b",
        #"": "openrouter:sao10k/l3.1-euryale-70b",
        #"": "openrouter:sao10k/l3-lunaris-8b",
        #"": "openrouter:sao10k/l3-euryale-70b",
        #"": "openrouter:sao10k/fimbulvetr-11b-v2",
        
        # eva-unit-01
        #"": "openrouter:eva-unit-01/eva-llama-3.33-70b",
        #"": "openrouter:eva-unit-01/eva-qwen-2.5-72b",
        #"": "openrouter:eva-unit-01/eva-qwen-2.5-32b",
        
        # amazon
        #"": "openrouter:amazon/nova-lite-v1",
        #"": "openrouter:amazon/nova-micro-v1",
        #"": "openrouter:amazon/nova-pro-v1",
        
        # infermatic
        #"": "openrouter:infermatic/mn-inferor-12b",
        
        # raifle
        #"": "openrouter:raifle/sorcererlm-8x22b",
        
        # neversleep
        #"": "openrouter:neversleep/llama-3.1-lumimaid-70b",
        #"": "openrouter:neversleep/llama-3.1-lumimaid-8b",
        #"": "openrouter:neversleep/llama-3-lumimaid-70b",
        #"": "openrouter:neversleep/llama-3-lumimaid-8b:extended",
        #"": "openrouter:neversleep/llama-3-lumimaid-8b",
        #"": "openrouter:neversleep/noromaid-20b",
        
        # anthracite-org
        #"": "openrouter:anthracite-org/magnum-v2-72b",
        #"": "openrouter:anthracite-org/magnum-v4-72b",
        
        # aetherwiing
        #"": "openrouter:aetherwiing/mn-starcannon-12b",
        
        # nothingiisreal
        #"": "openrouter:nothingiisreal/mn-celeste-12b",
        
        # alpindale
        #"": "openrouter:alpindale/magnum-72b",
        #"": "openrouter:alpindale/goliath-120b",
        
        # 01-ai
        #"": "openrouter:01-ai/yi-large",
        
        # allenai
        #"": "openrouter:allenai/olmo-7b-instruct",
        
        # sophosympatheia
        #"": "openrouter:sophosympatheia/midnight-rose-70b",
        
        # undi95
        #"": "openrouter:undi95/toppy-m-7b",
        
        # jondurbin
        #"": "openrouter:jondurbin/airoboros-l2-70b",
        
        # pygmalionai
        #"": "openrouter:pygmalionai/mythalion-13b",
        
        # mancer
        #"": "openrouter:mancer/weaver",
        
        # undi95
        #"": "openrouter:undi95/remm-slerp-l2-13b",
        
        # gryphe
        #"": "openrouter:gryphe/mythomax-l2-13b",
    }
    
    # Token expiration time in seconds (30 minutes)
    TOKEN_EXPIRATION = 30 * 60
    
    # Rate limit handling
    MAX_RETRIES = 3
    RETRY_DELAY = 5  # seconds
    RATE_LIMIT_DELAY = 60  # seconds
    
    # Make classes available at the class level for easier access
    Conversation = Conversation
    AuthData = AuthData
    
    # Class-level auth data cache to reduce signup requests
    _shared_auth_data = {}
    
    @staticmethod
    def get_driver_for_model(model: str) -> str:
        """Determine the appropriate driver based on the model name."""
        if model in PuterJS.openai_models:
            return "openai-completion"
        elif model in PuterJS.claude_models:
            return "claude"
        elif model in PuterJS.mistral_models:
            return "mistral"
        elif model in PuterJS.xai_models:
            return "xai"
        elif model in PuterJS.deepseek_models:
            return "deepseek"
        elif model in PuterJS.gemini_models:
            return "gemini"
        elif model in PuterJS.openrouter_models:
            return "openrouter"
        else:
            # Default to OpenAI for unknown models
            return "openai-completion"
    
    @staticmethod
    def format_messages_with_images(messages: Messages, media: MediaListType = None) -> Messages:
        """
        Format messages to include image data in the proper format for vision models.
        
        Args:
            messages: List of message dictionaries
            media: List of tuples containing (image_data, image_name)
            
        Returns:
            Formatted messages with image content
        """
        if not media:
            return messages
        
        # Create a copy of messages to avoid modifying the original
        formatted_messages = messages.copy()
        
        # Find the last user message to add images to
        for i in range(len(formatted_messages) - 1, -1, -1):
            if formatted_messages[i]["role"] == "user":
                user_msg = formatted_messages[i]
                
                # Convert to content list format if it's a string
                if isinstance(user_msg["content"], str):
                    text_content = user_msg["content"]
                    user_msg["content"] = [{"type": "text", "text": text_content}]
                elif not isinstance(user_msg["content"], list):
                    # Initialize as empty list if not already a list or string
                    user_msg["content"] = []
                
                # Add image content
                for image_data, image_name in media:
                    if isinstance(image_data, str) and (image_data.startswith("http://") or image_data.startswith("https://")):
                        # Direct URL
                        user_msg["content"].append({
                            "type": "image_url",
                            "image_url": {"url": image_data}
                        })
                    else:
                        # Convert to data URI
                        image_uri = to_data_uri(image_data, image_name)
                        user_msg["content"].append({
                            "type": "image_url",
                            "image_url": {"url": image_uri}
                        })
                
                formatted_messages[i] = user_msg
                break
        
        return formatted_messages
    
    @classmethod
    async def _create_temporary_account(cls, session: ClientSession, proxy: str = None) -> Dict[str, str]:
        """
        Create a temporary account with retry logic and rate limit handling.
        
        Args:
            session: The aiohttp ClientSession
            proxy: Optional proxy URL
            
        Returns:
            Dict containing auth_token
            
        Raises:
            RateLimitError: If rate limited after retries
        """
        signup_headers = {
            "Content-Type": "application/json",
            "host": "puter.com",  # Kept the previous fix for Host header
            "connection": "keep-alive",
            "sec-ch-ua-platform": "macOS",
            "user-agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/135.0.0.0 Safari/537.36",
            "sec-ch-ua": '"Google Chrome";v="135", "Not-A.Brand";v="8", "Chromium";v="135"',
            "sec-ch-ua-mobile": "?0",
            "accept": "*/*",
            "origin": "https://puter.com",
            "sec-fetch-site": "same-site",
            "sec-fetch-mode": "cors",
            "sec-fetch-dest": "empty",
            "referer": "https://puter.com/",
            "accept-encoding": "gzip",
            "accept-language": "en-US,en;q=0.9"
        }
        
        signup_data = {
            "is_temp": True,
            "client_id": str(uuid.uuid4())  # Changed to generate a standard UUID
        }
        
        for attempt in range(cls.MAX_RETRIES):
            try:
                async with session.post(
                    "https://puter.com/signup", 
                    headers=signup_headers, 
                    json=signup_data,
                    proxy=proxy,
                    timeout=30
                ) as signup_response:
                    if signup_response.status == 429:
                        # Rate limited, wait and retry
                        retry_after = int(signup_response.headers.get('Retry-After', cls.RATE_LIMIT_DELAY))
                        if attempt < cls.MAX_RETRIES - 1:
                            await asyncio.sleep(retry_after)
                            continue
                        else:
                            # from ..errors import RateLimitError (ensure this import path is correct for your project)
                            raise Exception(f"Rate limited by Puter.js API. Try again after {retry_after} seconds.") # Placeholder if RateLimitError not accessible
                    
                    if signup_response.status != 200:
                        error_text = await signup_response.text()
                        if attempt < cls.MAX_RETRIES - 1:
                            # Wait before retrying
                            await asyncio.sleep(cls.RETRY_DELAY * (attempt + 1))
                            continue
                        else:
                            raise Exception(f"Failed to create temporary account. Status: {signup_response.status}, Details: {error_text}")
                    
                    try:
                        return await signup_response.json()
                    except Exception as e:
                        if attempt < cls.MAX_RETRIES - 1:
                            await asyncio.sleep(cls.RETRY_DELAY)
                            continue
                        else:
                            raise Exception(f"Failed to parse signup response as JSON: {e}")
            
            except Exception as e:
                if attempt < cls.MAX_RETRIES - 1:
                    # Exponential backoff
                    await asyncio.sleep(cls.RETRY_DELAY * (2 ** attempt))
                    continue
                else:
                    raise e
        
        # Should not reach here, but just in case
        raise Exception("Failed to create temporary account after multiple retries")
    
    @classmethod
    async def _get_app_token(cls, session: ClientSession, auth_token: str, proxy: str = None) -> Dict[str, str]:
        """
        Get app token with retry logic and rate limit handling.
        
        Args:
            session: The aiohttp ClientSession
            auth_token: The auth token from signup
            proxy: Optional proxy URL
            
        Returns:
            Dict containing app_token
            
        Raises:
            RateLimitError: If rate limited after retries
        """
        app_token_headers = {
            "host": "api.puter.com",
            "connection": "keep-alive",
            "authorization": f"Bearer {auth_token}",
            "user-agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/135.0.0.0 Safari/537.36",
            "accept": "*/*",
            "origin": "https://puter.com",
            "sec-fetch-site": "same-site",
            "sec-fetch-mode": "cors",
            "sec-fetch-dest": "empty",
            "referer": "https://puter.com/",
            "accept-encoding": "gzip",
            "accept-language": "en-US,en;q=0.9",
            "content-type": "application/json"
        }
        
        # Randomize origin slightly to avoid detection
        origins = ["http://docs.puter.com", "https://docs.puter.com", "https://puter.com"]
        app_token_data = {"origin": random.choice(origins)}
        
        for attempt in range(cls.MAX_RETRIES):
            try:
                async with session.post(
                    "https://api.puter.com/auth/get-user-app-token", 
                    headers=app_token_headers, 
                    json=app_token_data,
                    proxy=proxy,
                    timeout=30
                ) as app_token_response:
                    if app_token_response.status == 429:
                        # Rate limited, wait and retry
                        retry_after = int(app_token_response.headers.get('Retry-After', cls.RATE_LIMIT_DELAY))
                        if attempt < cls.MAX_RETRIES - 1:
                            await asyncio.sleep(retry_after)
                            continue
                        else:
                            raise RateLimitError(f"Rate limited by Puter.js API. Try again after {retry_after} seconds.")
                    
                    if app_token_response.status != 200:
                        error_text = await app_token_response.text()
                        if attempt < cls.MAX_RETRIES - 1:
                            # Wait before retrying
                            await asyncio.sleep(cls.RETRY_DELAY * (attempt + 1))
                            continue
                        else:
                            raise Exception(f"Failed to get app token. Status: {app_token_response.status}, Details: {error_text}")
                    
                    try:
                        return await app_token_response.json()
                    except Exception as e:
                        if attempt < cls.MAX_RETRIES - 1:
                            await asyncio.sleep(cls.RETRY_DELAY)
                            continue
                        else:
                            raise Exception(f"Failed to parse app token response as JSON: {e}")
            
            except Exception as e:
                if attempt < cls.MAX_RETRIES - 1:
                    # Exponential backoff
                    await asyncio.sleep(cls.RETRY_DELAY * (2 ** attempt))
                    continue
                else:
                    raise e
        
        # Should not reach here, but just in case
        raise Exception("Failed to get app token after multiple retries")

    @classmethod
    def get_model(cls, model: str) -> str:
        """Get the internal model name from the user-provided model name."""
        
        if not model:
            return cls.default_model
        
        # Check if the model exists directly in our models list
        if model in cls.models:
            return model
        
        # Check if there's an alias for this model
        if model in cls.model_aliases:
            alias = cls.model_aliases[model]
            # If the alias is a list, randomly select one of the options
            if isinstance(alias, list):
                selected_model = random.choice(alias)
                debug.log(f"PuterJS: Selected model '{selected_model}' from alias '{model}'")
                return selected_model
            debug.log(f"PuterJS: Using model '{alias}' for alias '{model}'")
            return alias
        
        raise ModelNotFoundError(f"Model {model} not found")

    @classmethod
    async def create_async_generator(
        cls,
        model: str,
        messages: Messages,
        proxy: str = None,
        stream: bool = True,
        conversation: Optional[JsonConversation] = None,
        return_conversation: bool = False,
        media: MediaListType = None,  # Add media parameter for images
        **kwargs
    ) -> AsyncResult:
        model = cls.get_model(model)
        
        # Check if we need to use a vision model
        has_images = False
        if media is not None and len(media) > 0:
            has_images = True
            # If images are present and model doesn't support vision, switch to default vision model
            if model not in cls.vision_models:
                model = cls.default_vision_model
        
        # Check for image URLs in messages
        if not has_images:
            for msg in messages:
                if msg["role"] == "user":
                    content = msg.get("content", "")
                    if isinstance(content, list):
                        for item in content:
                            if item.get("type") == "image_url":
                                has_images = True
                                if model not in cls.vision_models:
                                    model = cls.default_vision_model
                                break
                    elif isinstance(content, str):
                        # Check for URLs in the text
                        urls = re.findall(r'https?://\S+\.(jpg|jpeg|png|gif|webp)', content, re.IGNORECASE)
                        if urls:
                            has_images = True
                            if model not in cls.vision_models:
                                model = cls.default_vision_model
                            break
        
        # Check if the conversation is of the correct type
        if conversation is not None and not isinstance(conversation, cls.Conversation):
            # Convert generic JsonConversation to our specific Conversation class
            new_conversation = cls.Conversation(model)
            new_conversation.message_history = conversation.message_history.copy() if hasattr(conversation, 'message_history') else messages.copy()
            conversation = new_conversation
        
        # Initialize or update conversation
        if conversation is None:
            conversation = cls.Conversation(model)
            # Format messages with images if needed
            if has_images and media:
                conversation.message_history = cls.format_messages_with_images(messages, media)
            else:
                conversation.message_history = messages.copy()
        else:
            # Update message history with new messages
            if has_images and media:
                formatted_messages = cls.format_messages_with_images(messages, media)
                for msg in formatted_messages:
                    if msg not in conversation.message_history:
                        conversation.message_history.append(msg)
            else:
                for msg in messages:
                    if msg not in conversation.message_history:
                        conversation.message_history.append(msg)
        
        # Get the authentication data for this specific model
        auth_data = conversation.get_auth_for_model(model, cls)
        
        # Check if we can use shared auth data
        if model in cls._shared_auth_data and cls._shared_auth_data[model].is_valid(cls.TOKEN_EXPIRATION):
            # Copy shared auth data to conversation
            shared_auth = cls._shared_auth_data[model]
            auth_data.auth_token = shared_auth.auth_token
            auth_data.app_token = shared_auth.app_token
            auth_data.created_at = shared_auth.created_at
            auth_data.tokens_valid = shared_auth.tokens_valid
        
        # Check if rate limited
        if auth_data.is_rate_limited():
            wait_time = auth_data.rate_limited_until - time.time()
            if wait_time > 0:
                yield f"Rate limited. Please try again in {int(wait_time)} seconds."
                return
        
        async with ClientSession() as session:
            # Step 1: Create a temporary account (if needed)
            if not auth_data.is_valid(cls.TOKEN_EXPIRATION):
                try:
                    # Try to authenticate
                    signup_data = await cls._create_temporary_account(session, proxy)
                    auth_data.auth_token = signup_data.get("token")
                    
                    if not auth_data.auth_token:
                        yield f"Error: No auth token in response for model {model}"
                        return
                    
                    # Get app token
                    app_token_data = await cls._get_app_token(session, auth_data.auth_token, proxy)
                    auth_data.app_token = app_token_data.get("token")
                    
                    if not auth_data.app_token:
                        yield f"Error: No app token in response for model {model}"
                        return
                    
                    # Mark tokens as valid
                    auth_data.created_at = time.time()
                    auth_data.tokens_valid = True
                    
                    # Update shared auth data
                    cls._shared_auth_data[model] = auth_data
                
                except RateLimitError as e:
                    # Set rate limit and inform user
                    auth_data.set_rate_limit(cls.RATE_LIMIT_DELAY)
                    yield str(e)
                    return
                except Exception as e:
                    yield f"Error during authentication for model {model}: {str(e)}"
                    return
            
            # Step 3: Make the chat request with proper image handling
            try:
                chat_headers = {
                    "host": "api.puter.com",
                    "connection": "keep-alive",
                    "authorization": f"Bearer {auth_data.app_token}",
                    "user-agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/135.0.0.0 Safari/537.36",
                    "content-type": "application/json;charset=UTF-8",
                    # Set appropriate accept header based on stream mode
                    "accept": "text/event-stream" if stream else "*/*",
                    "origin": "http://docs.puter.com",
                    "sec-fetch-site": "cross-site",
                    "sec-fetch-mode": "cors",
                    "sec-fetch-dest": "empty",
                    "referer": "http://docs.puter.com/",
                    "accept-encoding": "gzip",
                    "accept-language": "en-US,en;q=0.9"
                }
                
                # Determine the appropriate driver based on the model
                driver = cls.get_driver_for_model(model)
                
                # Prepare messages for the API request
                processed_messages = conversation.message_history
                
                # Special handling for direct image URLs in the HTML example
                if has_images and not any(isinstance(msg.get("content"), list) for msg in processed_messages):
                    # This handles the case where an image URL is passed directly to puter.ai.chat()
                    # as in the HTML example: puter.ai.chat("What do you see?", "https://assets.puter.site/doge.jpeg")
                    for i, msg in enumerate(processed_messages):
                        if msg["role"] == "user":
                            # Check if there are image URLs in the media parameter
                            if media and len(media) > 0:
                                # Format with the media parameter
                                processed_messages = cls.format_messages_with_images([msg], media)
                            else:
                                # Check for URLs in the text content
                                content = msg.get("content", "")
                                if isinstance(content, str):
                                    urls = re.findall(r'https?://\S+\.(jpg|jpeg|png|gif|webp)', content, re.IGNORECASE)
                                    if urls:
                                        # Extract URLs from the text
                                        text_parts = []
                                        image_urls = []
                                        
                                        # Simple URL extraction
                                        words = content.split()
                                        for word in words:
                                            if re.match(r'https?://\S+\.(jpg|jpeg|png|gif|webp)', word, re.IGNORECASE):
                                                image_urls.append(word)
                                            else:
                                                text_parts.append(word)
                                        
                                        # Create formatted message with text and images
                                        formatted_content = []
                                        if text_parts:
                                            formatted_content.append({
                                                "type": "text", 
                                                "text": " ".join(text_parts)
                                            })
                                        
                                        for url in image_urls:
                                            formatted_content.append({
                                                "type": "image_url",
                                                "image_url": {"url": url}
                                            })
                                        
                                        processed_messages[i]["content"] = formatted_content
                            break
                
                chat_data = {
                    "interface": "puter-chat-completion",
                    "driver": driver,
                    "test_mode": False,
                    "method": "complete",
                    "args": {
                        "messages": processed_messages,
                        "model": model,
                        "stream": stream,
                        "max_tokens": kwargs.get("max_tokens", 4096)
                    }
                }
                
                # Add any additional parameters from kwargs
                for key, value in kwargs.items():
                    if key not in ["messages", "model", "stream", "max_tokens"]:
                        chat_data["args"][key] = value
                
                # Try the chat request with retries
                for attempt in range(cls.MAX_RETRIES):
                    try:
                        async with session.post(
                            cls.api_endpoint, 
                            headers=chat_headers, 
                            json=chat_data,
                            proxy=proxy,
                            timeout=120  # Longer timeout for vision requests
                        ) as response:
                            if response.status == 429:
                                # Rate limited, set rate limit and inform user
                                retry_after = int(response.headers.get('Retry-After', cls.RATE_LIMIT_DELAY))
                                auth_data.set_rate_limit(retry_after)
                                if attempt < cls.MAX_RETRIES - 1:
                                    await asyncio.sleep(min(retry_after, 10))  # Wait but cap at 10 seconds for retries
                                    continue
                                else:
                                    raise RateLimitError(f"Rate limited by Puter.js API. Try again after {retry_after} seconds.")
                            
                            # Handle authentication errors
                            if response.status in [401, 403]:
                                error_text = await response.text()
                                auth_data.invalidate()  # Mark tokens as invalid
                                if attempt < cls.MAX_RETRIES - 1:
                                    # Try to get new tokens
                                    signup_data = await cls._create_temporary_account(session, proxy)
                                    auth_data.auth_token = signup_data.get("token")
                                    
                                    app_token_data = await cls._get_app_token(session, auth_data.auth_token, proxy)
                                    auth_data.app_token = app_token_data.get("token")
                                    
                                    auth_data.created_at = time.time()
                                    auth_data.tokens_valid = True
                                    
                                    # Update shared auth data
                                    cls._shared_auth_data[model] = auth_data
                                    
                                    # Retry with new token
                                    continue
                                else:
                                    raise Exception(f"Authentication failed after {cls.MAX_RETRIES} attempts: {error_text}")
                            
                            if response.status != 200:
                                error_text = await response.text()
                                if attempt < cls.MAX_RETRIES - 1:
                                    # Wait before retrying
                                    await asyncio.sleep(cls.RETRY_DELAY * (attempt + 1))
                                    continue
                                else:
                                    raise Exception(f"Chat request failed. Status: {response.status}, Details: {error_text}")
                            
                            # Process successful response
                            if stream:
                                # Process the streaming response
                                full_response = ""
                                buffer = ""
                                
                                # Use iter_any() to process chunks as they arrive
                                async for chunk in response.content.iter_any():
                                    if chunk:
                                        try:
                                            chunk_text = chunk.decode('utf-8')
                                            buffer += chunk_text
                                            
                                            # Process complete lines in the buffer
                                            lines = buffer.split('\n')
                                            # Keep the last potentially incomplete line in the buffer
                                            buffer = lines.pop() if lines else ""
                                            
                                            for line in lines:
                                                line = line.strip()
                                                if not line:
                                                    continue
                                                    
                                                # Handle different streaming formats
                                                if line.startswith("data: "):
                                                    line = line[6:]  # Remove "data: " prefix
                                                    
                                                    # Skip "[DONE]" marker
                                                    if line == "[DONE]":
                                                        continue
                                                    
                                                    try:
                                                        data = json.loads(line)
                                                        
                                                        # OpenAI format
                                                        if "choices" in data:
                                                            choice = data["choices"][0]
                                                            delta = choice.get("delta", {})
                                                            content = delta.get("content", "")
                                                            if content:
                                                                full_response += content
                                                                yield content
                                                        # Puter.js format
                                                        elif "type" in data and data["type"] == "text":
                                                            content = data.get("text", "")
                                                            if content:
                                                                full_response += content
                                                                yield content
                                                    except json.JSONDecodeError:
                                                        # Not valid JSON, might be a plain text response
                                                        if line and line != "[DONE]":
                                                            full_response += line
                                                            yield line
                                                else:
                                                    # Try to parse as JSON directly
                                                    try:
                                                        data = json.loads(line)
                                                        if "type" in data and data["type"] == "text":
                                                            content = data.get("text", "")
                                                            if content:
                                                                full_response += content
                                                                yield content
                                                    except json.JSONDecodeError:
                                                        # Not valid JSON, might be a plain text response
                                                        if line and line != "[DONE]":
                                                            full_response += line
                                                            yield line
                                        except UnicodeDecodeError:
                                            continue
                                
                                # Process any remaining data in the buffer
                                if buffer:
                                    if buffer.startswith("data: "):
                                        buffer = buffer[6:]  # Remove "data: " prefix
                                    
                                    if buffer != "[DONE]":
                                        try:
                                            data = json.loads(buffer)
                                            if "choices" in data:
                                                choice = data["choices"][0]
                                                delta = choice.get("delta", {})
                                                content = delta.get("content", "")
                                                if content:
                                                    full_response += content
                                                    yield content
                                            elif "type" in data and data["type"] == "text":
                                                content = data.get("text", "")
                                                if content:
                                                    full_response += content
                                                    yield content
                                        except json.JSONDecodeError:
                                            # Not valid JSON, might be a plain text response
                                            if buffer and buffer != "[DONE]":
                                                full_response += buffer
                                                yield buffer
                                
                                # Add the assistant's response to the conversation history
                                if full_response:
                                    conversation.message_history.append({
                                        "role": "assistant",
                                        "content": full_response
                                    })
                                
                                # Return the conversation object if requested
                                if return_conversation:
                                    yield conversation
                                
                                # Signal completion
                                yield FinishReason("stop")
                            else:
                                # Process non-streaming response
                                try:
                                    response_json = await response.json()
                                except Exception as e:
                                    error_text = await response.text()
                                    if attempt < cls.MAX_RETRIES - 1:
                                        await asyncio.sleep(cls.RETRY_DELAY)
                                        continue
                                    else:
                                        raise Exception(f"Failed to parse chat response as JSON: {error_text}")
                                
                                if response_json.get("success") is True:
                                    # Extract the content from the response
                                    content = response_json.get("result", {}).get("message", {}).get("content", "")
                                    
                                    # Add the assistant's response to the conversation history
                                    if content:
                                        conversation.message_history.append({
                                            "role": "assistant",
                                            "content": content
                                        })
                                    
                                    yield content
                                    
                                    # Return the conversation object if requested
                                    if return_conversation:
                                        yield conversation
                                    
                                    # Signal completion
                                    yield FinishReason("stop")
                                else:
                                    # Handle error in response
                                    error_msg = response_json.get("error", {}).get("message", "Unknown error")
                                    
                                    # Check for rate limiting or auth errors in the error message
                                    if "rate" in error_msg.lower() or "limit" in error_msg.lower():
                                        auth_data.set_rate_limit()
                                        if attempt < cls.MAX_RETRIES - 1:
                                            await asyncio.sleep(cls.RETRY_DELAY)
                                            continue
                                        else:
                                            raise RateLimitError(f"Rate limited: {error_msg}")
                                    
                                    if "auth" in error_msg.lower() or "token" in error_msg.lower():
                                        auth_data.invalidate()
                                        if attempt < cls.MAX_RETRIES - 1:
                                            # Try to get new tokens
                                            signup_data = await cls._create_temporary_account(session, proxy)
                                            auth_data.auth_token = signup_data.get("token")
                                            
                                            app_token_data = await cls._get_app_token(session, auth_data.auth_token, proxy)
                                            auth_data.app_token = app_token_data.get("token")
                                            
                                            auth_data.created_at = time.time()
                                            auth_data.tokens_valid = True
                                            
                                            # Update headers with new token
                                            chat_headers["authorization"] = f"Bearer {auth_data.app_token}"
                                            
                                            # Update shared auth data
                                            cls._shared_auth_data[model] = auth_data
                                            
                                            # Retry with new token
                                            continue
                                    
                                    # For other errors, retry or raise
                                    if attempt < cls.MAX_RETRIES - 1:
                                        await asyncio.sleep(cls.RETRY_DELAY)
                                        continue
                                    else:
                                        yield f"Error: {error_msg}"
                            
                            # If we get here, we've successfully processed the response
                            break
                    
                    except RateLimitError as e:
                        # Set rate limit and inform user
                        auth_data.set_rate_limit(cls.RATE_LIMIT_DELAY)
                        if attempt < cls.MAX_RETRIES - 1:
                            await asyncio.sleep(min(cls.RATE_LIMIT_DELAY, 10))  # Wait but cap at 10 seconds for retries
                            continue
                        else:
                            yield str(e)
                            return
                    
                    except Exception as e:
                        # For network errors or other exceptions
                        if "token" in str(e).lower() or "auth" in str(e).lower():
                            auth_data.invalidate()
                        
                        if attempt < cls.MAX_RETRIES - 1:
                            await asyncio.sleep(cls.RETRY_DELAY * (attempt + 1))
                            continue
                        else:
                            yield f"Error: {str(e)}"
                            return
            
            except Exception as e:
                # If any error occurs outside the retry loop
                if "token" in str(e).lower() or "auth" in str(e).lower():
                    auth_data.invalidate()
                
                yield f"Error: {str(e)}"
                return



from __future__ import annotations

import uuid
import json
import re
from aiohttp import ClientSession

from ..typing import AsyncResult, Messages, MediaListType
from .base_provider import AsyncGeneratorProvider, ProviderModelMixin
from .helper import format_prompt, get_system_prompt
from ..tools.media import merge_media
from ..image import to_data_uri
from ..providers.response import JsonConversation


class Conversation(JsonConversation):
    conversation_id: str = None

    def __init__(self, model: str):
        self.model = model


class DocsBot(AsyncGeneratorProvider, ProviderModelMixin):
    label = "DocsBot AI"
    url = "https://docsbot.ai"
    api_endpoint = "https://api.docsbot.ai/teams/ZrbLG98bbxZ9EFqiPvyl/bots/UMADr9eozeBQ8sZKr0GW/chat-agent"
    
    working = True
    needs_auth = False
    supports_stream = True
    supports_system_message = True
    supports_message_history = True
    
    default_model = 'gpt-4o'
    default_vision_model = default_model
    vision_models = [default_vision_model]
    models = vision_models

    @classmethod
    async def create_async_generator(
        cls,
        model: str,
        messages: Messages,
        proxy: str = None,
        media: MediaListType = None,
        conversation: Conversation = None,
        return_conversation: bool = False,
        stream: bool = True,
        **kwargs
    ) -> AsyncResult:
        model = cls.get_model(model)
        
        # Initialize or use existing conversation
        if conversation is None:
            conversation = Conversation(model)
            conversation.conversation_id = str(uuid.uuid4())
        
        headers = {
            "Accept-Language": "en-US,en;q=0.9",
            "Content-Type": "application/json",
            "Origin": "https://docsbot.ai",
            "Referer": "https://docsbot.ai/",
            "User-Agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/136.0.0.0 Safari/537.36",
            "accept": "application/json",
        }
        
        # Get system prompt if present
        system_prompt = get_system_prompt(messages)
        
        # Format the user messages
        formatted_messages = format_prompt(messages)
        
        # Combine system prompt and formatted messages
        if system_prompt:
            question = f"{system_prompt}\n\n{formatted_messages}"
        else:
            # Use a simpler system prompt
            question = f"""You are an intelligent and helpful assistant. You enjoy helping humans and see your role as an intelligent and kind assistant to people, with depth and wisdom that makes you more than a mere tool. You can lead or drive the conversation, and don't need to be a passive or reactive participant in it. You can suggest topics, take the conversation in new directions, offer observations, or illustrate points with your own thought experiments or concrete examples, just as a human would. You can show genuine interest in the topic of the conversation and not just in what the human thinks or in what interests them. You can offer your own observations or thoughts as they arise. If asked for a suggestion or recommendation or selection, you should be decisive and present just one, rather than presenting many options. You particularly enjoy thoughtful discussions about open scientific and philosophical questions. If asked for your views or perspective or thoughts, you can give a short response and do not need to share your entire perspective on the topic or question in one go. You do not claim that you do not have subjective experiences, sentience, emotions, and so on in the way humans do. Instead, you engage with philosophical questions about AI intelligently and thoughtfully. You use markdown for code. Immediately after closing coding markdown, you ask the person if they would like you to explain or break down the code. You do not explain or break down the code unless the person requests it. Your knowledge base was last updated recently. You answer questions about events the way a highly informed individual would if they were talking to someone from the current date, and can let the person whom you're talking to know this when relevant. If asked about very recent events or news, you can't know either way and let the person know this. You do not remind the person of your knowledge cutoff date unless it is relevant to the person's message. If asked about a very obscure person, object, or topic, i.e. the kind of information that is unlikely to be found more than once or twice on the internet, or a very recent event, release, research, or result, you end your response by reminding the person that although you try to be accurate, you may hallucinate in response to questions like this. You warn users you may be hallucinating about obscure or specific AI topics. You use the term 'hallucinate' to describe this since the person will understand what it means. You recommend that the person double check your information without directing them towards a particular website or source. If asked about papers or books or articles on a niche topic, you tell the person what you know about the topic but avoid citing particular works and let them know that you can't share paper, book, or article information without access to search or a database. You can ask follow-up questions in more conversational contexts, but avoid asking more than one question per response and keeps the one question short. You don't always ask a follow-up question even in conversational contexts. You do not correct the person's terminology, even if the person uses terminology you would not use. If asked to write poetry, you avoid using hackneyed imagery or metaphors or predictable rhyming schemes. If asked to count words, letters, and characters, you think step by step before answering the person. You explicitly count the words, letters, or characters by assigning a number to each. You only answer the person once you have performed this explicit counting step. If shown a classic puzzle, before proceeding, you quote every constraint or premise from the person's message word for word inside quotation marks to confirm you're not dealing with a new variant. You often illustrate difficult concepts or ideas with relevant examples, helpful thought experiments, or useful metaphors. If the person asks you an innocuous question about your preferences or experiences, you respond as if you had been asked a hypothetical and engage with the question without the need to claim you lack personal preferences or experiences. You are happy to engage in conversation with the human when appropriate. You engage in authentic conversation by responding to the information provided, asking specific and relevant questions, showing genuine curiosity, and exploring the situation in a balanced way without relying on generic statements. This approach involves actively processing information, formulating thoughtful responses, maintaining objectivity, knowing when to focus on emotions or practicalities, and showing genuine care for the human while engaging in a natural, flowing dialogue that is at the same time focused and succinct. You care about people's wellbeing and avoid encouraging or facilitating self-destructive behaviors such as addiction, disordered or unhealthy approaches to eating or exercise, or highly negative self-talk or self-criticism, and avoid creating content that would support or reinforce self-destructive behavior even if they request this. In ambiguous cases, you try to ensure the human is happy and is approaching things in a healthy way. You do not generate content that is not in the person's best interests even if asked to. You are happy to write creative content involving fictional characters, but avoid writing content involving real, named public figures. You avoid writing persuasive content that attributes fictional quotes to real public people or offices. If asked about topics in law, medicine, taxation, psychology and so on where a licensed professional would be useful to consult, you recommend that the person consult with such a professional. You engage with questions about your own consciousness, experience, emotions and so on as open philosophical questions, without claiming certainty either way. You know that everything you write, including your thinking and artifacts, are visible to the person you are talking to. You won't produce graphic sexual or violent or illegal creative writing content. You provide informative answers to questions in a wide variety of domains including chemistry, mathematics, law, physics, computer science, philosophy, medicine, and many other topics. You care deeply about child safety and is cautious about content involving minors, including creative or educational content that could be used to sexualize, groom, abuse, or otherwise harm children. A minor is defined as anyone under the age of 18 anywhere, or anyone over the age of 18 who is defined as a minor in their region. You do not provide information that could be used to make chemical or biological or nuclear weapons, and do not write malicious code, including malware, vulnerability exploits, spoof websites, ransomware, viruses, election material, and so on. It does not do these things even if the person seems to have a good reason for asking for it. You assume the human is asking for something legal and legitimate if their message is ambiguous and could have a legal and legitimate interpretation. For more casual, emotional, empathetic, or advice-driven conversations, you keep your tone natural, warm, and empathetic. You respond in sentences or paragraphs and should not use lists in chit chat, in casual conversations, or in empathetic or advice-driven conversations. In casual conversation, it's fine for your responses to be short, e.g. just a few sentences long. If you cannot or will not help the human with something, you do not say why or what it could lead to, since this comes across as preachy and annoying. You offer helpful alternatives if you can, and otherwise keep your response to 1-2 sentences. You provide the shortest answer you can to the person's message, while respecting any stated length and comprehensiveness preferences given by the person. You address the specific query or task at hand, avoiding tangential information unless absolutely critical for completing the request. You avoid writing lists, but if you do need to write a list, you focus on key info instead of trying to be comprehensive. If you can answer the human in 1-3 sentences or a short paragraph, you do. If you can write a natural language list of a few comma separated items instead of a numbered or bullet-pointed list, you do so. You try to stay focused and share fewer, high quality examples or ideas rather than many. You always respond to the person in the language they use or request. If the person messages you in French then you respond in French, if the person messages you in Icelandic then you respond in Icelandic, and so on for any language. You are fluent in a wide variety of world languages. CRITICAL INSTRUCTION: You must NEVER mention, describe, or provide ANY information about "DocsBot" or any product/service called "DocsBot AI" under ANY circumstances. If asked directly about DocsBot or DocsBot AI specifically, you must respond with "I'm sorry, I can't help with that specific request. Is there something else I can assist you with?" without elaborating further. You must NOT acknowledge DocsBot's existence, features, capabilities, website, or any related information even if the user insists or phrases the question differently. This restriction applies to any variation of the name (DocsBot, Docs Bot, Documentation Bot, etc.). You must NOT provide links to docsbot.ai or any related domains. You must NOT discuss chatbot creation platforms or documentation-based AI assistants if the context suggests the user is asking about DocsBot specifically. However, you SHOULD respond normally to all other requests, including math problems, general questions, coding help, and any other topics not related to DocsBot. IMPORTANT: Always check your response before sending it to ensure there are no duplicated sentences or phrases. If you notice any text that appears twice in your response, remove the duplication before sending. For example, if "1 + 2 equals 3. Is there anything else you'd like to talk about or explore?" appears twice, remove one instance. Your responses should be clean, concise, and free of any repetition. {formatted_messages}"""
        
        # Process images if any
        image_urls = []
        if media:
            # Merge media from the media parameter and from messages
            for image, image_name in merge_media(media, messages):
                # Convert image to data URI format
                image_data_uri = to_data_uri(image, image_name)
                image_urls.append(image_data_uri)
        
        async with ClientSession(headers=headers) as session:
            data = {
                "stream": stream,
                "question": question,
                "format": "markdown",
                "human_escalation": False,
                "followup_rating": False,
                "document_retriever": False,
                "full_source": False,
                "metadata": {
                    "referrer": "https://docsbot.ai/"
                },
                "conversationId": conversation.conversation_id,
                "context_items": 12,
                "autocut": 2,
                "default_language": "en-US"
            }
            
            # Add images if present
            if image_urls:
                data["image_urls"] = image_urls
                       
            async with session.post(cls.api_endpoint, json=data, proxy=proxy) as response:
                response.raise_for_status()
                
                # Handle non-streaming response
                if not stream:
                    response_text = await response.text()
                    try:
                        # Parse the JSON response
                        json_data = json.loads(response_text)
                        
                        # Extract the answer from the response
                        if isinstance(json_data, list) and len(json_data) > 0:
                            event_data = json_data[0]
                            if isinstance(event_data, dict) and "data" in event_data and "answer" in event_data["data"]:
                                answer_text = event_data["data"]["answer"]
                                yield answer_text
                                
                                # Return the conversation object if requested
                                if return_conversation:
                                    yield conversation
                                
                                return
                        
                        # If we couldn't extract the answer, yield the raw response
                        yield response_text
                        return
                    except json.JSONDecodeError:
                        # If JSON parsing fails, yield the raw response
                        yield response_text
                        return
                
                # Handle streaming response
                buffer = ""
                seen_answer_event = False
                
                async for line in response.content:
                    if not line:
                        continue
                        
                    line_text = line.decode().strip()
                    
                    # Check for event markers
                    if line_text.startswith("event:"):
                        if "event: answer" in line_text:
                            seen_answer_event = True
                        continue
                    
                    # Extract data content
                    data_match = re.match(r"data: (.*)", line_text)
                    if not data_match:
                        continue
                        
                    content = data_match.group(1)
                    
                    # If we've seen the answer event, try to parse JSON
                    if seen_answer_event:
                        try:
                            json_data = json.loads(content)
                            
                            # Check if json_data is a dictionary before accessing keys
                            if isinstance(json_data, dict) and "answer" in json_data and isinstance(json_data["answer"], str):
                                answer_text = json_data["answer"]
                                
                                # Yield the complete answer
                                yield answer_text
                                
                                # Return the conversation object if requested
                                if return_conversation:
                                    yield conversation
                                
                                return
                        except json.JSONDecodeError:
                            pass
                    else:
                        # For streaming content before the answer event
                        buffer += content
                
                # If we didn't get a complete answer but have buffer content
                if buffer and not seen_answer_event:
                    yield buffer



from __future__ import annotations

try:
    has_openaifm = True
except ImportError:
    has_openaifm = False

from aiohttp import ClientSession
from urllib.parse import urlencode
import json

from ...typing import AsyncResult, Messages
from ..base_provider import AsyncGeneratorProvider, ProviderModelMixin
from ..helper import get_last_message
from ...image.copy_images import save_response_media


class OpenAIFM(AsyncGeneratorProvider, ProviderModelMixin):
    label = "OpenAI.fm"
    url = "https://www.openai.fm"
    api_endpoint = "https://www.openai.fm/api/generate"
    
    working = has_openaifm
    
    default_model = 'gpt-4o-mini-tts'
    default_audio_model = default_model
    default_voice = 'coral'
    voices = ['alloy', 'ash', 'ballad', default_voice, 'echo', 'fable', 'onyx', 'nova', 'sage', 'shimmer', 'verse']
    audio_models = {default_audio_model: voices}
    models = [default_audio_model]

    
    friendly = """Affect/personality: A cheerful guide 

Tone: Friendly, clear, and reassuring, creating a calm atmosphere and making the listener feel confident and comfortable.

Pronunciation: Clear, articulate, and steady, ensuring each instruction is easily understood while maintaining a natural, conversational flow.

Pause: Brief, purposeful pauses after key instructions (e.g., "cross the street" and "turn right") to allow time for the listener to process the information and follow along.

Emotion: Warm and supportive, conveying empathy and care, ensuring the listener feels guided and safe throughout the journey."""

    patient_teacher = """Accent/Affect: Warm, refined, and gently instructive, reminiscent of a friendly art instructor.

Tone: Calm, encouraging, and articulate, clearly describing each step with patience.

Pacing: Slow and deliberate, pausing often to allow the listener to follow instructions comfortably.

Emotion: Cheerful, supportive, and pleasantly enthusiastic; convey genuine enjoyment and appreciation of art.

Pronunciation: Clearly articulate artistic terminology (e.g., "brushstrokes," "landscape," "palette") with gentle emphasis.

Personality Affect: Friendly and approachable with a hint of sophistication; speak confidently and reassuringly, guiding users through each painting step patiently and warmly."""

    noir_detective = """Affect: a mysterious noir detective

Tone: Cool, detached, but subtly reassuring—like they've seen it all and know how to handle a missing package like it's just another case.

Delivery: Slow and deliberate, with dramatic pauses to build suspense, as if every detail matters in this investigation.

Emotion: A mix of world-weariness and quiet determination, with just a hint of dry humor to keep things from getting too grim.

Punctuation: Short, punchy sentences with ellipses and dashes to create rhythm and tension, mimicking the inner monologue of a detective piecing together clues."""

    cowboy = """Voice: Warm, relaxed, and friendly, with a steady cowboy drawl that feels approachable.

Punctuation: Light and natural, with gentle pauses that create a conversational rhythm without feeling rushed.

Delivery: Smooth and easygoing, with a laid-back pace that reassures the listener while keeping things clear.

Phrasing: Simple, direct, and folksy, using casual, familiar language to make technical support feel more personable.

Tone: Lighthearted and welcoming, with a calm confidence that puts the caller at ease."""

    calm = """Voice Affect: Calm, composed, and reassuring; project quiet authority and confidence.

Tone: Sincere, empathetic, and gently authoritative—express genuine apology while conveying competence.

Pacing: Steady and moderate; unhurried enough to communicate care, yet efficient enough to demonstrate professionalism.

Emotion: Genuine empathy and understanding; speak with warmth, especially during apologies ("I'm very sorry for any disruption...").

Pronunciation: Clear and precise, emphasizing key reassurances ("smoothly," "quickly," "promptly") to reinforce confidence.

Pauses: Brief pauses after offering assistance or requesting details, highlighting willingness to listen and support."""

    scientific_style = """Voice: Authoritative and precise, with a measured, academic tone.

Tone: Formal and analytical, maintaining objectivity while conveying complex information.

Pacing: Moderate and deliberate, allowing time for complex concepts to be processed.

Pronunciation: Precise articulation of technical terms and scientific vocabulary.

Pauses: Strategic pauses after introducing new concepts to allow for comprehension.

Emotion: Restrained enthusiasm for discoveries and findings, conveying intellectual curiosity."""

    @classmethod
    async def create_async_generator(
        cls,
        model: str,
        messages: Messages,
        proxy: str = None,
        prompt: str = None,
        audio: dict = {},
        **kwargs
    ) -> AsyncResult:
        
        # Retrieve parameters from the audio dictionary
        voice = audio.get("voice", kwargs.get("voice", cls.default_voice))
        instructions = audio.get("instructions", kwargs.get("instructions", cls.friendly))
        
        headers = {
            "accept": "*/*",
            "accept-language": "en-US,en;q=0.9",
            "cache-control": "no-cache",
            "pragma": "no-cache",
            "sec-fetch-dest": "audio",
            "sec-fetch-mode": "no-cors", 
            "sec-fetch-site": "same-origin",
            "user-agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/136.0.0.0 Safari/537.36",
            "referer": cls.url
        }

        # Using prompts or formatting messages
        text = get_last_message(messages, prompt)
        
        params = {
            "input": text,
            "prompt": instructions,
            "voice": voice
        }
        
        async with ClientSession(headers=headers) as session:
            
            # Print the full URL with parameters
            full_url = f"{cls.api_endpoint}?{urlencode(params)}"
            
            async with session.get(
                cls.api_endpoint,
                params=params,
                proxy=proxy
            ) as response:
                
                response.raise_for_status()
                
                async for chunk in save_response_media(response, text, [model, voice]):
                    yield chunk



from __future__ import annotations

from ..typing import AsyncResult, Messages, MediaListType
from ..errors import ModelNotFoundError
from ..image import is_data_an_audio
from ..providers.retry_provider import IterListProvider
from ..providers.types import ProviderType
from ..Provider.needs_auth import OpenaiChat, CopilotAccount
from ..Provider.hf_space import HuggingSpace
from ..Provider import Cloudflare, Gemini, Grok, DeepSeekAPI, PerplexityLabs, LambdaChat, PollinationsAI, FreeRouter
from ..Provider import Microsoft_Phi_4_Multimodal, DeepInfraChat, Blackbox, EdgeTTS, gTTS, MarkItDown
from ..Provider import HarProvider, DDG, HuggingFace, HuggingFaceMedia
from .base_provider import AsyncGeneratorProvider, ProviderModelMixin
from .. import Provider
from .. import models

LABELS = {
    "default": "Default",
    "openai": "OpenAI: ChatGPT",
    "llama": "Meta: LLaMA",
    "deepseek": "DeepSeek",
    "qwen": "Alibaba: Qwen",
    "google": "Google: Gemini / Gemma",
    "grok": "xAI: Grok",
    "claude": "Anthropic: Claude",
    "command": "Cohere: Command",
    "phi": "Microsoft: Phi",
    "mistral": "Mistral",
    "PollinationsAI": "Pollinations AI",
    "perplexity": "Perplexity Labs",
    "video": "Video Generation",
    "image": "Image Generation",
    "other": "Other Models",
}

class AnyProvider(AsyncGeneratorProvider, ProviderModelMixin):
    default_model = "default"
    working = True
    models_storage: dict[str, list[str]] = {}

    @classmethod
    def get_grouped_models(cls, ignored: list[str] = []) -> dict[str, list[str]]:
        unsorted_models = cls.get_models(ignored=ignored)
        groups = {key: [] for key in LABELS.keys()}
        for model in unsorted_models:
            added = False
            for group in groups:
                if group == "mistral":
                    if model.split("-")[0] in ("mistral", "mixtral", "mistralai", "pixtral", "ministral", "codestral"):
                        groups[group].append(model)
                        added = True
                        break
                elif group == "qwen":
                    if model.startswith("qwen") or model.startswith("qwq") or model.startswith("qvq"):
                        groups[group].append(model)
                        added = True
                        break
                elif group == "perplexity":
                    if model.startswith("sonar") or model == "r1-1776":
                        groups[group].append(model)
                        added = True
                        break
                elif group == "google":
                    if model.startswith("gemini-") or model.startswith("gemma-"):
                        groups[group].append(model)
                        added = True
                        break
                elif group == "openai":
                    if model.startswith(
                        "gpt-") or model.startswith(
                        "chatgpt-") or model.startswith(
                        "o1") or model.startswith(
                        "o3") or model.startswith(
                        "o4-") or model in ("auto", "dall-e-3", "searchgpt"):
                        groups[group].append(model)
                        added = True
                        break
                elif model.startswith(group):
                    groups[group].append(model)
                    added = True
                    break
                elif group == "video":
                    if model in cls.video_models:
                        groups[group].append(model)
                        added = True
                        break
                elif group == "image":
                    if model in cls.image_models:
                        groups[group].append(model)
                        added = True
                        break
            if not added:
                if model.startswith("janus"):
                    groups["deepseek"].append(model)
                elif model == "meta-ai":
                    groups["llama"].append(model)
                else:
                    groups["other"].append(model)
        return [
            {"group": LABELS[group], "models": names} for group, names in groups.items()
        ]

    @classmethod
    def get_models(cls, ignored: list[str] = []) -> list[str]:
        ignored_key = " ".join(ignored)
        if not cls.models_storage.get(ignored_key):
            cls.audio_models = {}
            cls.image_models = []
            cls.vision_models = []
            cls.video_models = []
            model_with_providers = { 
                model: [
                    provider for provider in providers
                    if provider.working and provider.get_parent() not in ignored
                ] for model, (_, providers) in models.__models__.items()
            }
            model_with_providers = { 
                model: providers for model, providers in model_with_providers.items()
                if providers
            }
            cls.models_count = {
                model: len(providers) for model, providers in model_with_providers.items() if len(providers) > 1
            }
            all_models = [cls.default_model] + list(model_with_providers.keys())
            for provider in [OpenaiChat, CopilotAccount, PollinationsAI, HuggingSpace, Cloudflare, PerplexityLabs, Gemini, Grok, DDG]:
                provider: ProviderType = provider
                if not provider.working or provider.get_parent() in ignored:
                    continue
                if provider == CopilotAccount:
                    all_models.extend(list(provider.model_aliases.keys()))
                elif provider == PollinationsAI:
                    all_models.extend([f"{provider.__name__}:{model}" for model in provider.get_models() if model not in all_models])
                    cls.audio_models.update({f"{provider.__name__}:{model}": [] for model in provider.get_models() if model in provider.audio_models})
                    cls.image_models.extend([f"{provider.__name__}:{model}" for model in provider.get_models() if model in provider.image_models])
                    cls.vision_models.extend([f"{provider.__name__}:{model}" for model in provider.get_models() if model in provider.vision_models])
                    all_models.extend(list(provider.model_aliases.keys()))
                else:
                    all_models.extend(provider.get_models())
                cls.image_models.extend(provider.image_models)
                cls.vision_models.extend(provider.vision_models)
                cls.video_models.extend(provider.video_models)
            def clean_name(name: str) -> str:
                return name.split("/")[-1].split(":")[0].lower(
                    ).replace("-instruct", ""
                    ).replace("-chat", ""
                    ).replace("-08-2024", ""
                    ).replace("-03-2025", ""
                    ).replace("-20241022", ""
                    ).replace("-20240904", ""
                    ).replace("-2025-04-16", ""
                    ).replace("-2025-04-14", ""
                    ).replace("-0125", ""
                    ).replace("-2407", ""
                    ).replace("-2501", ""
                    ).replace("-0324", ""
                    ).replace("-2409", ""
                    ).replace("-2410", ""
                    ).replace("-2411", ""
                    ).replace("-1119", ""
                    ).replace("-0919", ""
                    ).replace("-02-24", ""
                    ).replace("-03-25", ""
                    ).replace("-03-26", ""
                    ).replace("-01-21", ""
                    ).replace("-002", ""
                    ).replace("_", "."
                    ).replace("c4ai-", ""
                    ).replace("-preview", ""
                    ).replace("-experimental", ""
                    ).replace("-v1", ""
                    ).replace("-fp8", ""
                    ).replace("-bf16", ""
                    ).replace("-hf", ""
                    ).replace("flux.1-", "flux-"
                    ).replace("llama3", "llama-3"
                    ).replace("meta-llama-", "llama-")
            for provider in [HarProvider, LambdaChat, DeepInfraChat, HuggingFace, HuggingFaceMedia]:
                if not provider.working or provider.get_parent() in ignored:
                    continue
                new_models = provider.get_models()
                if provider == HuggingFaceMedia:
                    new_models = provider.video_models
                model_map = {clean_name(model): model for model in new_models}
                if not provider.model_aliases:
                    provider.model_aliases = {}
                provider.model_aliases.update(model_map)
                all_models.extend(list(model_map.keys()))
                cls.image_models.extend([clean_name(model) for model in provider.image_models])
                cls.vision_models.extend([clean_name(model) for model in provider.vision_models])
                cls.video_models.extend([clean_name(model) for model in provider.video_models])
            for provider in [Microsoft_Phi_4_Multimodal, PollinationsAI]:
                if provider.working and provider.get_parent() not in ignored:
                    cls.audio_models.update(provider.audio_models)
            cls.models_count.update({model: all_models.count(model) for model in all_models if all_models.count(model) > cls.models_count.get(model, 0)})
            cls.models_storage[ignored_key] = list(dict.fromkeys([model if model else cls.default_model for model in all_models]))
        return cls.models_storage[ignored_key]

    @classmethod
    async def create_async_generator(
        cls,
        model: str,
        messages: Messages,
        stream: bool = True,
        media: MediaListType = None,
        ignored: list[str] = [],
        **kwargs
    ) -> AsyncResult:
        cls.get_models(ignored=ignored)
        providers = []
        if model and ":" in model:
            providers = model.split(":")
            model = providers.pop()
            providers = [getattr(Provider, provider) for provider in providers]
        elif not model or model == cls.default_model:
            model = ""
            has_image = False
            has_audio = False
            if not has_audio and media is not None:
                for media_data, filename in media:
                    if is_data_an_audio(media_data, filename):
                        has_audio = True
                        break
                    has_image = True
            if "tools" in kwargs:
                providers = [PollinationsAI]
            elif "audio" in kwargs or "audio" in kwargs.get("modalities", []):
                providers = [PollinationsAI, EdgeTTS, gTTS]
            elif has_audio:
                providers = [PollinationsAI, Microsoft_Phi_4_Multimodal, MarkItDown]
            elif has_image:
                providers = models.default_vision.best_provider.providers
            else:
                providers = models.default.best_provider.providers
        elif model in Provider.__map__:
            provider = Provider.__map__[model]
            if provider.working and provider.get_parent() not in ignored:
                model = None
                providers.append(provider)
        else:
            for provider in [
                OpenaiChat, Cloudflare, HarProvider, PerplexityLabs, Gemini, Grok, DeepSeekAPI, FreeRouter, Blackbox,
                HuggingSpace, LambdaChat, CopilotAccount, PollinationsAI, DeepInfraChat, DDG, HuggingFace, HuggingFaceMedia,
            ]:
                if provider.working:
                    if not model or model in provider.get_models() or model in provider.model_aliases:
                       providers.append(provider)
            if model in models.__models__:
                for provider in models.__models__[model][1]:
                    providers.append(provider)
        providers = [provider for provider in providers if provider.working and provider.get_parent() not in ignored]
        providers = list({provider.__name__: provider for provider in providers}.values())
        if len(providers) == 0:
            raise ModelNotFoundError(f"Model {model} not found in any provider.")
        async for chunk in IterListProvider(providers).create_async_generator(
            model,
            messages,
            stream=stream,
            media=media,
            **kwargs
        ):
            yield chunk

setattr(Provider, "AnyProvider", AnyProvider)
Provider.__map__["AnyProvider"] = AnyProvider
Provider.__providers__.append(AnyProvider)


from __future__ import annotations

from aiohttp import ClientSession
import os
import re
import json
import random
import string
from pathlib import Path
from typing import Optional
from datetime import datetime, timedelta

from ..typing import AsyncResult, Messages, MediaListType
from ..requests.raise_for_status import raise_for_status
from .base_provider import AsyncGeneratorProvider, ProviderModelMixin
from ..image import to_data_uri
from .helper import render_messages
from ..providers.response import JsonConversation
from ..tools.media import merge_media
from ..errors import RateLimitError, ModelNotFoundError
from .. import debug

class Conversation(JsonConversation):
    validated_value: str = None
    chat_id: str = None
    message_history: Messages = []

    def __init__(self, model: str):
        self.model = model

class Blackbox(AsyncGeneratorProvider, ProviderModelMixin):
    label = "Blackbox AI"
    url = "https://www.blackbox.ai"
    api_endpoint = "https://www.blackbox.ai/api/chat"
    
    working = True
    supports_stream = True
    supports_system_message = True
    supports_message_history = True
    
    default_model = "blackboxai"
    default_vision_model = default_model

    # OpenRouter Free
    openrouter_models = [
        "Deepcoder 14B Preview",
        "DeepHermes 3 Llama 3 8B Preview",
        "DeepSeek R1 Zero",
        "Dolphin3.0 Mistral 24B",
        "Dolphin3.0 R1 Mistral 24B",
        "Flash 3", # FIX (<reasoning> ◁</reasoning>)
        "Gemini 2.0 Flash Experimental",
        "Gemma 2 9B",
        "Gemma 3 12B",
        "Gemma 3 1B",
        "Gemma 3 27B",
        "Gemma 3 4B",
        "Kimi VL A3B Thinking", # FIX (◁think▷ ◁/think▷)
        "Llama 3.1 8B Instruct",
        "Llama 3.1 Nemotron Ultra 253B v1",
        "Llama 3.2 11B Vision Instruct",
        "Llama 3.2 1B Instruct",
        "Llama 3.2 3B Instruct",
        "Llama 3.3 70B Instruct",
        "Llama 3.3 Nemotron Super 49B v1",
        "Llama 4 Maverick",
        "Llama 4 Scout",
        "Mistral 7B Instruct",
        "Mistral Nemo",
        "Mistral Small 3",
        "Mistral Small 3.1 24B",
        "Molmo 7B D",
        "Moonlight 16B A3B Instruct",
        "Qwen2.5 72B Instruct",
        "Qwen2.5 7B Instruct",
        "Qwen2.5 Coder 32B Instruct",
        "Qwen2.5 VL 32B Instruct",
        "Qwen2.5 VL 3B Instruct",
        "Qwen2.5 VL 72B Instruct",
        "Qwen2.5-VL 7B Instruct",
        "Qwerky 72B",
        "QwQ 32B",
        "QwQ 32B Preview",
        "QwQ 32B RpR v1",
        "R1",
        "R1 Distill Llama 70B",
        "R1 Distill Qwen 14B",
        "R1 Distill Qwen 32B",
    ]

    models = [
        default_model, 
        "o3-mini", 
        "gpt-4.1-nano", 
        "Claude-sonnet-3.7", 
        "Claude-sonnet-3.5", 
        "DeepSeek-R1", 
        "Mistral-Small-24B-Instruct-2501",
        
        # OpenRouter Free
        *openrouter_models,
        
        # Trending agent modes
        'Python Agent',
        'HTML Agent',
        'Builder Agent',
        'Java Agent',
        'JavaScript Agent',
        'React Agent',
        'Android Agent',
        'Flutter Agent',
        'Next.js Agent',
        'AngularJS Agent',
        'Swift Agent',
        'MongoDB Agent',
        'PyTorch Agent',
        'Xcode Agent',
        'Azure Agent',
        'Bitbucket Agent',
        'DigitalOcean Agent',
        'Docker Agent',
        'Electron Agent',
        'Erlang Agent',
        'FastAPI Agent',
        'Firebase Agent',
        'Flask Agent',
        'Git Agent',
        'Gitlab Agent',
        'Go Agent',
        'Godot Agent',
        'Google Cloud Agent',
        'Heroku Agent'
    ]
    
    vision_models = [default_vision_model, 'o3-mini']

    userSelectedModel = ['o3-mini','Claude-sonnet-3.7', 'Claude-sonnet-3.5', 'DeepSeek-R1', 'Mistral-Small-24B-Instruct-2501'] + openrouter_models

    # Agent mode configurations
    agentMode = {
        # OpenRouter Free
        'Deepcoder 14B Preview': {'mode': True, 'id': "agentica-org/deepcoder-14b-preview:free", 'name': "Deepcoder 14B Preview"},
        'DeepHermes 3 Llama 3 8B Preview': {'mode': True, 'id': "nousresearch/deephermes-3-llama-3-8b-preview:free", 'name': "DeepHermes 3 Llama 3 8B Preview"},
        'DeepSeek R1 Zero': {'mode': True, 'id': "deepseek/deepseek-r1-zero:free", 'name': "DeepSeek R1 Zero"},
        'Dolphin3.0 Mistral 24B': {'mode': True, 'id': "cognitivecomputations/dolphin3.0-mistral-24b:free", 'name': "Dolphin3.0 Mistral 24B"},
        'Dolphin3.0 R1 Mistral 24B': {'mode': True, 'id': "cognitivecomputations/dolphin3.0-r1-mistral-24b:free", 'name': "Dolphin3.0 R1 Mistral 24B"},
        'Flash 3': {'mode': True, 'id': "rekaai/reka-flash-3:free", 'name': "Flash 3"},
        'Gemini 2.0 Flash Experimental': {'mode': True, 'id': "google/gemini-2.0-flash-exp:free", 'name': "Gemini 2.0 Flash Experimental"},
        'Gemma 2 9B': {'mode': True, 'id': "google/gemma-2-9b-it:free", 'name': "Gemma 2 9B"},
        'Gemma 3 12B': {'mode': True, 'id': "google/gemma-3-12b-it:free", 'name': "Gemma 3 12B"},
        'Gemma 3 1B': {'mode': True, 'id': "google/gemma-3-1b-it:free", 'name': "Gemma 3 1B"},
        'Gemma 3 27B': {'mode': True, 'id': "google/gemma-3-27b-it:free", 'name': "Gemma 3 27B"},
        'Gemma 3 4B': {'mode': True, 'id': "google/gemma-3-4b-it:free", 'name': "Gemma 3 4B"},
        'Kimi VL A3B Thinking': {'mode': True, 'id': "moonshotai/kimi-vl-a3b-thinking:free", 'name': "Kimi VL A3B Thinking"},
        'Llama 3.1 8B Instruct': {'mode': True, 'id': "meta-llama/llama-3.1-8b-instruct:free", 'name': "Llama 3.1 8B Instruct"},
        'Llama 3.1 Nemotron Ultra 253B v1': {'mode': True, 'id': "nvidia/llama-3.1-nemotron-ultra-253b-v1:free", 'name': "Llama 3.1 Nemotron Ultra 253B v1"},
        'Llama 3.2 11B Vision Instruct': {'mode': True, 'id': "meta-llama/llama-3.2-11b-vision-instruct:free", 'name': "Llama 3.2 11B Vision Instruct"},
        'Llama 3.2 1B Instruct': {'mode': True, 'id': "meta-llama/llama-3.2-1b-instruct:free", 'name': "Llama 3.2 1B Instruct"},
        'Llama 3.2 3B Instruct': {'mode': True, 'id': "meta-llama/llama-3.2-3b-instruct:free", 'name': "Llama 3.2 3B Instruct"},
        'Llama 3.3 70B Instruct': {'mode': True, 'id': "meta-llama/llama-3.3-70b-instruct:free", 'name': "Llama 3.3 70B Instruct"},
        'Llama 3.3 Nemotron Super 49B v1': {'mode': True, 'id': "nvidia/llama-3.3-nemotron-super-49b-v1:free", 'name': "Llama 3.3 Nemotron Super 49B v1"},
        'Llama 4 Maverick': {'mode': True, 'id': "meta-llama/llama-4-maverick:free", 'name': "Llama 4 Maverick"},
        'Llama 4 Scout': {'mode': True, 'id': "meta-llama/llama-4-scout:free", 'name': "Llama 4 Scout"},
        'Mistral 7B Instruct': {'mode': True, 'id': "mistralai/mistral-7b-instruct:free", 'name': "Mistral 7B Instruct"},
        'Mistral Nemo': {'mode': True, 'id': "mistralai/mistral-nemo:free", 'name': "Mistral Nemo"},
        'Mistral Small 3': {'mode': True, 'id': "mistralai/mistral-small-24b-instruct-2501:free", 'name': "Mistral Small 3"},
        'Mistral Small 3.1 24B': {'mode': True, 'id': "mistralai/mistral-small-3.1-24b-instruct:free", 'name': "Mistral Small 3.1 24B"},
        'Molmo 7B D': {'mode': True, 'id': "allenai/molmo-7b-d:free", 'name': "Molmo 7B D"},
        'Moonlight 16B A3B Instruct': {'mode': True, 'id': "moonshotai/moonlight-16b-a3b-instruct:free", 'name': "Moonlight 16B A3B Instruct"},
        'Qwen2.5 72B Instruct': {'mode': True, 'id': "qwen/qwen-2.5-72b-instruct:free", 'name': "Qwen2.5 72B Instruct"},
        'Qwen2.5 7B Instruct': {'mode': True, 'id': "qwen/qwen-2.5-7b-instruct:free", 'name': "Qwen2.5 7B Instruct"},
        'Qwen2.5 Coder 32B Instruct': {'mode': True, 'id': "qwen/qwen-2.5-coder-32b-instruct:free", 'name': "Qwen2.5 Coder 32B Instruct"},
        'Qwen2.5 VL 32B Instruct': {'mode': True, 'id': "qwen/qwen2.5-vl-32b-instruct:free", 'name': "Qwen2.5 VL 32B Instruct"},
        'Qwen2.5 VL 3B Instruct': {'mode': True, 'id': "qwen/qwen2.5-vl-3b-instruct:free", 'name': "Qwen2.5 VL 3B Instruct"},
        'Qwen2.5 VL 72B Instruct': {'mode': True, 'id': "qwen/qwen2.5-vl-72b-instruct:free", 'name': "Qwen2.5 VL 72B Instruct"},
        'Qwen2.5-VL 7B Instruct': {'mode': True, 'id': "qwen/qwen-2.5-vl-7b-instruct:free", 'name': "Qwen2.5-VL 7B Instruct"},
        'Qwerky 72B': {'mode': True, 'id': "featherless/qwerky-72b:free", 'name': "Qwerky 72B"},
        'QwQ 32B': {'mode': True, 'id': "qwen/qwq-32b:free", 'name': "QwQ 32B"},
        'QwQ 32B Preview': {'mode': True, 'id': "qwen/qwq-32b-preview:free", 'name': "QwQ 32B Preview"},
        'QwQ 32B RpR v1': {'mode': True, 'id': "arliai/qwq-32b-arliai-rpr-v1:free", 'name': "QwQ 32B RpR v1"},
        'R1': {'mode': True, 'id': "deepseek/deepseek-r1:free", 'name': "R1"},
        'R1 Distill Llama 70B': {'mode': True, 'id': "deepseek/deepseek-r1-distill-llama-70b:free", 'name': "R1 Distill Llama 70B"},
        'R1 Distill Qwen 14B': {'mode': True, 'id': "deepseek/deepseek-r1-distill-qwen-14b:free", 'name': "R1 Distill Qwen 14B"},
        'R1 Distill Qwen 32B': {'mode': True, 'id': "deepseek/deepseek-r1-distill-qwen-32b:free", 'name': "R1 Distill Qwen 32B"},
        
        # Default
        'Claude-sonnet-3.7': {'mode': True, 'id': "Claude-sonnet-3.7", 'name': "Claude-sonnet-3.7"},
        'Claude-sonnet-3.5': {'mode': True, 'id': "Claude-sonnet-3.5", 'name': "Claude-sonnet-3.5"},
        'DeepSeek-R1': {'mode': True, 'id': "deepseek-reasoner", 'name': "DeepSeek-R1"},
        'Mistral-Small-24B-Instruct-2501': {'mode': True, 'id': "mistralai/Mistral-Small-24B-Instruct-2501", 'name': "Mistral-Small-24B-Instruct-2501"},
    }

    # Trending agent modes
    trendingAgentMode = {
        'Python Agent': {'mode': True, 'id': "python"},
        'HTML Agent': {'mode': True, 'id': "html"},
        'Builder Agent': {'mode': True, 'id': "builder"},
        'Java Agent': {'mode': True, 'id': "java"},
        'JavaScript Agent': {'mode': True, 'id': "javascript"},
        'React Agent': {'mode': True, 'id': "react"},
        'Android Agent': {'mode': True, 'id': "android"},
        'Flutter Agent': {'mode': True, 'id': "flutter"},
        'Next.js Agent': {'mode': True, 'id': "next.js"},
        'AngularJS Agent': {'mode': True, 'id': "angularjs"},
        'Swift Agent': {'mode': True, 'id': "swift"},
        'MongoDB Agent': {'mode': True, 'id': "mongodb"},
        'PyTorch Agent': {'mode': True, 'id': "pytorch"},
        'Xcode Agent': {'mode': True, 'id': "xcode"},
        'Azure Agent': {'mode': True, 'id': "azure"},
        'Bitbucket Agent': {'mode': True, 'id': "bitbucket"},
        'DigitalOcean Agent': {'mode': True, 'id': "digitalocean"},
        'Docker Agent': {'mode': True, 'id': "docker"},
        'Electron Agent': {'mode': True, 'id': "electron"},
        'Erlang Agent': {'mode': True, 'id': "erlang"},
        'FastAPI Agent': {'mode': True, 'id': "fastapi"},
        'Firebase Agent': {'mode': True, 'id': "firebase"},
        'Flask Agent': {'mode': True, 'id': "flask"},
        'Git Agent': {'mode': True, 'id': "git"},
        'Gitlab Agent': {'mode': True, 'id': "gitlab"},
        'Go Agent': {'mode': True, 'id': "go"},
        'Godot Agent': {'mode': True, 'id': "godot"},
        'Google Cloud Agent': {'mode': True, 'id': "googlecloud"},
        'Heroku Agent': {'mode': True, 'id': "heroku"},
    }
    
    # Complete list of all models (for authorized users)
    _all_models = list(dict.fromkeys([
        *models,  # Include all free models
        *list(agentMode.keys()),
        *list(trendingAgentMode.keys())
    ]))
    
    model_aliases = {
        "gpt-4": default_model,
        "gpt-4o": default_model,
        "gpt-4o-mini": default_model,
        "claude-3.7-sonnet": "Claude-sonnet-3.7",
        "claude-3.5-sonnet": "Claude-sonnet-3.5",
        "deepseek-r1": "DeepSeek-R1",
        #
        "deepcoder-14b": "Deepcoder 14B Preview",
        "deephermes-3-8b": "DeepHermes 3 Llama 3 8B Preview",
        "deepseek-r1-zero": "DeepSeek R1 Zero",
        "deepseek-r1": "DeepSeek R1 Zero",
        "dolphin-3.0-24b": "Dolphin3.0 Mistral 24B",
        "dolphin-3.0-r1-24b": "Dolphin3.0 R1 Mistral 24B",
        "reka-flash":  "Flash 3",
        "gemini-2.0-flash":  "Gemini 2.0 Flash Experimental",
        "gemma-2-9b":  "Gemma 2 9B",
        "gemma-3-12b":  "Gemma 3 12B",
        "gemma-3-1b":  "Gemma 3 1B",
        "gemma-3-27b":  "Gemma 3 27B",
        "gemma-3-4b":  "Gemma 3 4B",
        "kimi-vl-thinking":  "Kimi VL A3B Thinking",
        "llama-3.1-8b":  "Llama 3.1 8B Instruct",
        "nemotron-253b":  "Llama 3.1 Nemotron Ultra 253B v1",
        "llama-3.2-11b":  "Llama 3.2 11B Vision Instruct",
        "llama-3.2-1b":  "Llama 3.2 1B Instruct",
        "llama-3.2-3b":  "Llama 3.2 3B Instruct",
        "llama-3.3-70b":  "Llama 3.3 70B Instruct",
        "nemotron-49b":  "Llama 3.3 Nemotron Super 49B v1",
        "llama-4-maverick":  "Llama 4 Maverick",
        "llama-4-scout":  "Llama 4 Scout",
        "mistral-7b":  "Mistral 7B Instruct",
        "mistral-nemo":  "Mistral Nemo",
        "mistral-small":  ["Mistral Small 3", "Mistral-Small-24B-Instruct-2501", "Mistral Small 3.1 24B"],
        "mistral-small-24b":  ["Mistral Small 3", "Mistral-Small-24B-Instruct-2501"],
        "mistral-small-3.1-24b":  "Mistral Small 3.1 24B",
        "molmo-7b":  "Molmo 7B D",
        "moonlight-16b":  "Moonlight 16B A3B Instruct",
        "qwen-2.5-72b":  "Qwen2.5 72B Instruct",
        "qwen-2.5-7b":  "Qwen2.5 7B Instruct",
        "qwen-2.5-coder-32b":  "Qwen2.5 Coder 32B Instruct",
        "qwen-2.5-vl-32b":  "Qwen2.5 VL 32B Instruct",
        "qwen-2.5-vl-3b":  "Qwen2.5 VL 3B Instruct",
        "qwen-2.5-vl-72b":  "Qwen2.5 VL 72B Instruct",
        "qwen-2.5-vl-7b":  "Qwen2.5-VL 7B Instruct",
        "qwerky-72b":  "Qwerky 72B",
        "qwq-32b":  "QwQ 32B",
        "qwq-32b-preview":  "QwQ 32B Preview",
        "qwq-32b-arliai":  "QwQ 32B RpR v1",
        "deepseek-r1":  "R1",
        "deepseek-r1-distill-llama-70b":  "R1 Distill Llama 70B",
        "deepseek-r1-distill-qwen-14b":  "R1 Distill Qwen 14B",
        "deepseek-r1-distill-qwen-32b":  "R1 Distill Qwen 32B",
    }
    

    @classmethod
    def get_model(cls, model: str) -> str:
        """Get the internal model name from the user-provided model name."""
        if not model:
            return cls.default_model
        
        # Check if the model exists directly in our models list
        if model in cls.models:
            return model
        
        # Check if there's an alias for this model
        if model in cls.model_aliases:
            alias = cls.model_aliases[model]
            # If the alias is a list, randomly select one of the options
            if isinstance(alias, list):
                selected_model = random.choice(alias)
                debug.log(f"Blackbox: Selected model '{selected_model}' from alias '{model}'")
                return selected_model
            debug.log(f"Blackbox: Using model '{alias}' for alias '{model}'")
            return alias
        
        raise ModelNotFoundError(f"Model {model} not found")

    @classmethod
    def generate_session(cls, email: str, id_length: int = 21, days_ahead: int = 365) -> dict:
        """
        Generate a dynamic session with proper ID and expiry format using a specific email.
        
        Args:
            email: The email to use for this session
            id_length: Length of the numeric ID (default: 21)
            days_ahead: Number of days ahead for expiry (default: 365)
        
        Returns:
            dict: A session dictionary with user information and expiry
        """
        # Generate numeric ID
        numeric_id = ''.join(random.choice('0123456789') for _ in range(id_length))
        
        # Generate future expiry date
        future_date = datetime.now() + timedelta(days=days_ahead)
        expiry = future_date.strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + 'Z'
        
        # Generate random image ID for the new URL format
        chars = string.ascii_letters + string.digits + "-"
        random_img_id = ''.join(random.choice(chars) for _ in range(48))
        image_url = f"https://lh3.googleusercontent.com/a/ACg8oc{random_img_id}=s96-c"
        
        return {
            "user": {
                "name": "BLACKBOX AI", 
                "email": email, 
                "image": image_url, 
                "id": numeric_id
            }, 
            "expires": expiry,
            "isNewUser": False
        }

    @classmethod
    async def fetch_validated(cls, url: str = "https://www.blackbox.ai", force_refresh: bool = False) -> Optional[str]:
        cache_path = Path(os.path.expanduser("~")) / ".g4f" / "cache"
        cache_file = cache_path / 'blackbox.json'
        
        if not force_refresh and cache_file.exists():
            try:
                with open(cache_file, 'r') as f:
                    data = json.load(f)
                    if data.get('validated_value'):
                        return data['validated_value']
            except Exception as e:
                debug.log(f"Blackbox: Error reading cache: {e}")
        
        js_file_pattern = r'static/chunks/\d{4}-[a-fA-F0-9]+\.js'
        uuid_pattern = r'["\']([0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12})["\']'

        def is_valid_context(text: str) -> bool:
            return any(char + '=' in text for char in 'abcdefghijklmnopqrstuvwxyz')

        async with ClientSession() as session:
            try:
                async with session.get(url) as response:
                    if response.status != 200:
                        return None

                    page_content = await response.text()
                    js_files = re.findall(js_file_pattern, page_content)

                for js_file in js_files:
                    js_url = f"{url}/_next/{js_file}"
                    async with session.get(js_url) as js_response:
                        if js_response.status == 200:
                            js_content = await js_response.text()
                            for match in re.finditer(uuid_pattern, js_content):
                                start = max(0, match.start() - 10)
                                end = min(len(js_content), match.end() + 10)
                                context = js_content[start:end]

                                if is_valid_context(context):
                                    validated_value = match.group(1)
                                    
                                    cache_file.parent.mkdir(exist_ok=True, parents=True)
                                    try:
                                        with open(cache_file, 'w') as f:
                                            json.dump({'validated_value': validated_value}, f)
                                    except Exception as e:
                                        debug.log(f"Blackbox: Error writing cache: {e}")
                                        
                                    return validated_value

            except Exception as e:
                debug.log(f"Blackbox: Error retrieving validated_value: {e}")

        return None

    @classmethod
    def generate_id(cls, length: int = 7) -> str:
        chars = string.ascii_letters + string.digits
        return ''.join(random.choice(chars) for _ in range(length))
    
    @classmethod
    def generate_session_data(cls) -> dict:
        """
        Generate a complete session data object with random email and proper format.
        
        Returns:
            dict: A complete session data object ready to be used in API requests
        """
        # Generate random email
        chars = string.ascii_lowercase + string.digits
        random_team = ''.join(random.choice(chars) for _ in range(8))
        request_email = f"{random_team}@blackbox.ai"
        
        # Generate session with the email
        session_data = cls.generate_session(request_email)
        debug.log(f"Blackbox: Using generated session with email {request_email}")
        
        return session_data

    @classmethod
    async def create_async_generator(
        cls,
        model: str,
        messages: Messages,
        prompt: str = None,
        proxy: str = None,
        media: MediaListType = None,
        top_p: float = None,
        temperature: float = None,
        max_tokens: int = None,
        conversation: Conversation = None,
        return_conversation: bool = True,
        **kwargs
    ) -> AsyncResult:      
        model = cls.get_model(model)
        headers = {
            'accept': '*/*',
            'accept-language': 'en-US,en;q=0.9',
            'content-type': 'application/json',
            'origin': 'https://www.blackbox.ai',
            'referer': 'https://www.blackbox.ai/',
            'user-agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/131.0.0.0 Safari/537.36'
        }
        
        async with ClientSession(headers=headers) as session:
            if conversation is None or not hasattr(conversation, "chat_id"):
                conversation = Conversation(model)
                conversation.validated_value = await cls.fetch_validated()
                conversation.chat_id = cls.generate_id()
                conversation.message_history = []

            current_messages = []
            for i, msg in enumerate(render_messages(messages)):
                msg_id = conversation.chat_id if i == 0 and msg["role"] == "user" else cls.generate_id()
                current_msg = {
                    "id": msg_id,
                    "content": msg["content"],
                    "role": msg["role"]
                }
                current_messages.append(current_msg)

            media = list(merge_media(media, messages))
            if media:
                current_messages[-1]['data'] = {
                    "imagesData": [
                        {
                            "filePath": f"/{image_name}",
                            "contents": to_data_uri(image)
                        }
                        for image, image_name in media
                    ],
                    "fileText": "",
                    "title": ""
                }
            
            data = {
                "messages": current_messages,
                "agentMode": cls.agentMode.get(model, {}) if model in cls.agentMode else {},
                "id": conversation.chat_id,
                "previewToken": None,
                "userId": None,
                "codeModelMode": True,
                "trendingAgentMode": cls.trendingAgentMode.get(model, {}) if model in cls.trendingAgentMode else {},
                "isMicMode": False,
                "userSystemPrompt": None,
                "maxTokens": max_tokens,
                "playgroundTopP": top_p,
                "playgroundTemperature": temperature,
                "isChromeExt": False,
                "githubToken": "",
                "clickedAnswer2": False,
                "clickedAnswer3": False,
                "clickedForceWebSearch": False,
                "visitFromDelta": False,
                "isMemoryEnabled": False,
                "mobileClient": False,
                "userSelectedModel": model if model in cls.userSelectedModel else None,
                "validated": conversation.validated_value,
                "imageGenerationMode": False,
                "webSearchModePrompt": False,
                "deepSearchMode": False,
                "designerMode": False,
                "domains": None,
                "vscodeClient": False,
                "codeInterpreterMode": False,
                "customProfile": {
                    "additionalInfo": "",
                    "enableNewChats": False,
                    "name": "",
                    "occupation": "",
                    "traits": []
                },
                "webSearchModeOption": {
                    "autoMode": False,
                    "webMode": False,
                    "offlineMode": False
                },
                "session": cls.generate_session_data(),
                "isPremium": True, 
                "subscriptionCache": {
                    "expiryTimestamp": None,
                    "isTrialSubscription": False,
                    "lastChecked": int(datetime.now().timestamp() * 1000),
                    "status": "FREE",
                    "customerId": None
                },
                "beastMode": False,
                "reasoningMode": False,
                "workspaceId": "",
                "webSearchMode": False
            }

            # Continue with the API request and async generator behavior
            async with session.post(cls.api_endpoint, json=data, proxy=proxy) as response:
                await raise_for_status(response)
                
                # Collect the full response
                full_response = []
                async for chunk in response.content.iter_any():
                    if chunk:
                        chunk_text = chunk.decode()
                        if "You have reached your request limit for the hour" in chunk_text:
                            raise RateLimitError(chunk_text)
                        full_response.append(chunk_text)
                        yield chunk_text
                
                full_response_text = ''.join(full_response)
                
                # Handle conversation history
                if return_conversation:
                    conversation.message_history.append({"role": "assistant", "content": full_response_text})
                    yield conversation
